# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Apply an All-Air template HVAC to a list of Honeybee Rooms.
_
All-air systems provide both ventilation and satisfaction of heating + cooling
demand with the same stream of warm/cool air. As such, they often grant tight
control over zone humidity. However, because such systems often involve the
cooling of air only to reheat it again, they are often more energy intensive
than systems that separate ventilation from the meeting of thermal loads.
-

    Args:
        _rooms: Honeybee Rooms to which the input template HVAC will be assigned.
            This can also be a Honeybee Model for which all conditioned Rooms
            will be assigned the HVAC system.
        _system_type: Text for the specific type of all-air system and equipment.
            The "HB All-Air HVAC Templates" component has a full list of the
            supported all-air system templates.
        _vintage_: Text for the vintage of the template system. This will be used
            to set efficiencies for various pieces of equipment within the system.
            The "HB Building Vintages" component has a full list of supported
            HVAC vintages. (Default: ASHRAE_2019).
        _name_: Text to set the name for the HVAC system and to be incorporated into
            unique HVAC identifier. If the name is not provided, a random name
            will be assigned.
        _economizer_: Text to indicate the type of air-side economizer used on the
            HVAC system. Economizers will mix in a greater amount of outdoor
            air to cool the zone (rather than running the cooling system) when
            the zone needs cooling and the outdoor air is cooler than the zone.
            Choose from the options below. (Default: NoEconomizer).
                * NoEconomizer
                * DifferentialDryBulb
                * DifferentialEnthalpy
                * DifferentialDryBulbAndEnthalpy
                * FixedDryBulb
                * FixedEnthalpy
                * ElectronicEnthalpy
        sensible_hr_: A number between 0 and 1 for the effectiveness of sensible
            heat recovery within the system. Typical values range from 0.5 for
            simple glycol loops to 0.81 for enthalpy wheels (the latter tends to
            be fiarly expensive for air-based systems). (Default: 0).
        latent_hr_: A number between 0 and 1 for the effectiveness of latent heat
            recovery within the system. Typical values are 0 for all types of
            heat recovery except enthalpy wheels, which can have values as high
            as 0.76. (Default: 0).
        dcv_: Boolean to note whether demand controlled ventilation should be
            used on the system, which will vary the amount of ventilation air
            according to the occupancy schedule of the zone. (Default: False).

    Returns:
        rooms: The input Rooms with an all-air HVAC system applied.
"""

ghenv.Component.Name = "HB All-Air HVAC"
ghenv.Component.NickName = 'AllAirHVAC'
ghenv.Component.Message = '1.2.3'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '4 :: HVAC'
ghenv.Component.AdditionalHelpFromDocStrings = '2'

import os
import json

try:  # import the honeybee extension
    from honeybee.typing import clean_and_id_ep_string, clean_ep_string
    from honeybee.model import Model
    from honeybee.room import Room
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:  # import the honeybee-energy extension
    from honeybee_energy.config import folders
    from honeybee_energy.hvac.allair import EQUIPMENT_TYPES_DICT
    from honeybee_energy.lib.schedules import schedule_by_identifier
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))

# dictionary to get correct vintages
vintages = {
    'DOE_Ref_Pre_1980': 'DOE_Ref_Pre_1980',
    'DOE_Ref_1980_2004': 'DOE_Ref_1980_2004',
    'ASHRAE_2004': 'ASHRAE_2004',
    'ASHRAE_2007': 'ASHRAE_2007',
    'ASHRAE_2010': 'ASHRAE_2010',
    'ASHRAE_2013': 'ASHRAE_2013',
    'ASHRAE_2016': 'ASHRAE_2016',
    'ASHRAE_2019': 'ASHRAE_2019',
    'DOE Ref Pre-1980': 'DOE_Ref_Pre_1980',
    'DOE Ref 1980-2004': 'DOE_Ref_1980_2004',
    '90.1-2004': 'ASHRAE_2004',
    '90.1-2007': 'ASHRAE_2007',
    '90.1-2010': 'ASHRAE_2010',
    '90.1-2013': 'ASHRAE_2013',
    'pre_1980': 'DOE_Ref_Pre_1980',
    '1980_2004': 'DOE_Ref_1980_2004',
    '2004': 'ASHRAE_2004',
    '2007': 'ASHRAE_2007',
    '2010': 'ASHRAE_2010',
    '2013': 'ASHRAE_2013',
    '2016': 'ASHRAE_2016',
    '2019': 'ASHRAE_2019',
    None: 'ASHRAE_2019'
}

# dictionary of HVAC template names
ext_folder = folders.standards_extension_folders[0]
hvac_reg = os.path.join(ext_folder, 'hvac_registry.json')
with open(hvac_reg, 'r') as f:
    hvac_dict = json.load(f)


if all_required_inputs(ghenv.Component):
    # extract any rooms from input Models and duplicate the rooms
    rooms = []
    for hb_obj in _rooms:
        if isinstance(hb_obj, Model):
            rooms.extend([room.duplicate() for room in hb_obj.rooms])
        elif isinstance(hb_obj, Room):
            rooms.append(hb_obj.duplicate())
        else:
            raise ValueError(
                'Expected Honeybee Room or Model. Got {}.'.format(type(hb_obj)))

    # process any input properties for the HVAC system
    try:  # get the class for the HVAC system
        try:
            _sys_name = hvac_dict[_system_type]
        except KeyError:
            _sys_name = _system_type
        hvac_class = EQUIPMENT_TYPES_DICT[_sys_name]
    except KeyError:
        raise ValueError('System Type "{}" is not recognized as an all-air HVAC '
                         'system.'.format(_system_type))
    vintage = vintages[_vintage_]  # get the vintage of the HVAC
    name = clean_and_id_ep_string('All-Air HVAC') if _name_ is None else clean_ep_string(_name_)

    # create the HVAC
    hvac = hvac_class(name, vintage, _sys_name)
    if _economizer_ is not None:
        hvac.economizer_type = _economizer_
    if sensible_hr_ is not None:
        hvac.sensible_heat_recovery = sensible_hr_
    if latent_hr_ is not None:
        hvac.latent_heat_recovery = latent_hr_
    if dcv_:
        hvac.demand_controlled_ventilation = True
    if _name_ is not None:
        hvac.display_name = _name_

    # apply the HVAC system to the rooms
    for room in rooms:
        if room.properties.energy.is_conditioned:
            room.properties.energy.hvac = hvac



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Create a window material to describe a single glass pane corresponding to a
layer in a window construction. This material can be plugged into the "HB Window
Construction" component.
-

    Args:
        _name: Text to set the name for the material and to be incorporated into
            a unique material identifier.
        _thickness: Number for the thickness of the glass layer [m]. Typical values
            range from 0.003 meters (3 mm) to 0.012 meters (12 mm).
        _transmittance_: Number between 0 and 1 for the transmittance of both solar
            radiation and visible light through the glass at normal
            incidence. (Default: 0.85 for clear uncoated glass).
        _reflectance_: Number between 0 and 1 for the reflectance of both solar
            radiation and visible light off of the front side of the glass at
            normal incidence. (Default: 0.075 for clear uncoated glass).
        _t_infrared_: Long-wave transmittance of the glass at normal
            incidence. (Default: 0).
        _emiss_front_: Number between 0 and 1 for the infrared hemispherical
            emissivity of the front side of the glass. (Defaul: 0.84, which
            is typical of clear glass).
        _emiss_back_: Number between 0 and 1 for the infrared hemispherical
            emissivity of the back side of the glass. (Default: 0.84, which
            is typical of clear glass).
        _conductivity_: Number for the thermal conductivity of the glass in
            W/m-K. (Default: 0.9, whih is typical of clear glass).

    Returns:
        mat: A window material that describes a single glass pane and can be
            assigned to a Honeybee Window construction.
"""

ghenv.Component.Name = 'HB Glass Material'
ghenv.Component.NickName = 'GlassMat'
ghenv.Component.Message = '1.2.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '1 :: Constructions'
ghenv.Component.AdditionalHelpFromDocStrings = '6'

try:  # import the core honeybee dependencies
    from honeybee.typing import clean_and_id_ep_string, clean_ep_string
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:  # import the honeybee-energy dependencies
    from honeybee_energy.material.glazing import EnergyWindowMaterialGlazing
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:  # import ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # set the default material properties
    _transmittance_ = 0.85 if _transmittance_ is None else _transmittance_
    _reflectance_ = 0.075 if _reflectance_ is None else _reflectance_
    _t_infrared_ = 0 if _t_infrared_ is None else _t_infrared_
    _emiss_front_ = 0.84 if _emiss_front_ is None else _emiss_front_
    _emiss_back_ = 0.84 if _emiss_back_ is None else _emiss_back_
    _conductivity_ = 0.9 if _conductivity_ is None else _conductivity_
    name = clean_and_id_ep_string('GlassMaterial') if _name_ is None else \
        clean_ep_string(_name_)

    # create the material
    mat = EnergyWindowMaterialGlazing(
        name, _thickness, _transmittance_, _reflectance_,
        _transmittance_, _reflectance_, _t_infrared_, _emiss_front_, _emiss_back_,
        _conductivity_)
    if _name_ is not None:
        mat.display_name = _name_



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Deconstruct a ServiceHotWater object into its constituient properties.
-

    Args:
        _hot_water: A ServiceHotWater object to be deconstructed.

    Returns:
        name_: Text string for the service hot water display name.
         flow_per_area: A numerical value for the total volume flow rate of water
            per unit area of floor (L/h-m2).
        schedule: A fractional schedule for the use of hot water over the course of
            the year. The fractional values will get multiplied by the
            _flow_per_area to yield a complete water usage profile.
        target_temp: The target temperature of the water out of the tap in Celsius.
            This the temperature after the hot water has been mixed with cold
            water from the water mains.
        sensible_fract: A number between 0 and 1 for the fraction of the total
            hot water load given off as sensible heat in the zone.
        latent_fract: A number between 0 and 1 for the fraction of the total
            hot water load that is latent (as opposed to sensible).
"""

ghenv.Component.Name = 'HB Deconstruct Hot Water'
ghenv.Component.NickName = 'DecnstrHotWater'
ghenv.Component.Message = '1.2.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '3 :: Loads'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:
    from honeybee_energy.load.hotwater import ServiceHotWater
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component) and _hot_water is not None:
    # check the input
    assert isinstance(_hot_water, ServiceHotWater), \
        'Expected ServiceHotWater object. Got {}.'.format(type(_hot_water))

    # get the properties of the object
    name = _hot_water.display_name
    flow_per_area = _hot_water.flow_per_area
    schedule = _hot_water.schedule
    target_temp = _hot_water.target_temperature
    sensible_fract = _hot_water.sensible_fraction
    latent_fract = _hot_water.latent_fraction



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Create a custom gas gap material that corresponds to a layer in a window construction.
This material can be plugged into the "HB Window Construction" component.
_
This object allows you to specify specific values for conductivity,
viscosity and specific heat through the following formula:
    property = A + (B * T)
where:
    A, B = regression coefficients for the gas
    T = temperature [K]
-

    Args:
        _name_: Text to set the name for the material and to be incorporated into
            a unique material identifier.
        _thickness: Number for the thickness of the gas gap layer in meters.
        _conductivity_a: First conductivity coefficient.
            Or condictivity in [W/m-K] if b coefficient is 0.
        _viscosity_a: First viscosity coefficient.
            Or viscosity in [kg/m-s] if b coefficient is 0.
        _specific_heat_a: First specific heat coefficient.
            Or the specific heat in [J/kg-K] if b coefficient is 0.
        _conductivity_b_: Second conductivity coefficient. (Default: 0).
        _viscosity_b_: Second viscosity coefficient. (Default: 0).
        _specific_heat_b_: Second specific heat coefficient. (Default: 0).
        _spec_heat_ratio_: A number for the the ratio of the specific heat a contant
            pressure, to the specific heat at constant volume. (Default: 1.0 for Air).
        _mol_weight_: Number between 20 and 200 for the mass of 1 mol of
            the substance in grams. (Default: 20.0).

    Returns:
        mat: A custom gas gap material that describes a layer in a window construction
            and can be assigned to a Honeybee Window construction.
"""

ghenv.Component.Name = 'HB Custom Window Gap Material'
ghenv.Component.NickName = 'CustomGapMat'
ghenv.Component.Message = '1.2.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '1 :: Constructions'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:  # import the core honeybee dependencies
    from honeybee.typing import clean_and_id_ep_string, clean_ep_string
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:  # import the honeybee-energy dependencies
    from honeybee_energy.material.gas import EnergyWindowMaterialGasCustom
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:  # import ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # set the default material properties
    _conductivity_b_ = 0 if _conductivity_b_ is None else _conductivity_b_
    _viscosity_b_ = 0 if _viscosity_b_ is None else _viscosity_b_
    _specific_heat_b_ = 0 if _specific_heat_b_ is None else _specific_heat_b_
    _spec_heat_ratio_ = 1.0 if _spec_heat_ratio_ is None else _spec_heat_ratio_
    _mol_weight_ = 20.0 if _mol_weight_ is None else _mol_weight_
    name = clean_and_id_ep_string('GapMaterial') if _name_ is None else \
        clean_ep_string(_name_)

    # set the non-exposed inputs
    _conductivity_c_, _viscosity_c_, _specific_heat_c_ = 0, 0, 0

    # create the material
    mat = EnergyWindowMaterialGasCustom(
        name, _thickness,
        _conductivity_a, _viscosity_a, _specific_heat_a,
        _conductivity_b_, _viscosity_b_, _specific_heat_b_,
        _conductivity_c_, _viscosity_c_, _specific_heat_c_,
        _spec_heat_ratio_, _mol_weight_)
    if _name_ is not None:
        mat.display_name = _name_



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Modify a native Grasshopper Gene Pool component into a format that makes it easy
to input daily schedules into the other schedule components.
_
This essentially turns the Gene Pool into a list of 24 sliders representing an hourly
timeseries over a day. Each slider runs from the _low_bound_ to _high_bound_.
-

    Args:
        _gene_pool: The output of a native Grasshopper Gene Pool component.
        _template_: An integer for the template for which the values of the gene
            pool will be set. Default: 0. Choose one of the templates:
            -
            0 - maximum values from 9:00 to 17:00
            1 - minimum values from 9:00 to 17:00
            2 - maximum values from 7:00 to 22:00
            3 - minimum values from 7:00 to 22:00
            4 - maximum values from 0:00 to 24:00
            5 - minimum values from 0:00 to 24:00
        _low_bound_: A number for the lower boundary of the schedule values.
            Defaut: 0.
        _up_bound_: A number for the upper boundary of the schedule range.
            Default: 1.
        _decimals_: An integer greater than or equal to 0 for the number of
            decimal places to use in each slider. Default: 1.
        _run: Set to "True" to run the component and modify the connected gene pool.
            Note that you on't be able to edit a connected gene pool while this
            input is set to "True".
    
    Returns:
        report: Reports, errors, warnings, etc.
"""

ghenv.Component.Name = "HB Gene Pool to Day Schedule"
ghenv.Component.NickName = 'GenePoolToDaySch'
ghenv.Component.Message = '1.2.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '2 :: Schedules'
ghenv.Component.AdditionalHelpFromDocStrings = "1"

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


# template schedules
templates = [
    [0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0],
    [1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1],
    [0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0],
    [1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1],
    [1] * 24,
    [0] * 24
]


if all_required_inputs(ghenv.Component) and _run:
    # set the default values
    _template_ = 0 if _template_ is None else _template_
    _low_bound_ = 0.0 if _low_bound_ is None else _low_bound_
    _up_bound_ = 1.0 if _up_bound_ is None else _up_bound_
    _decimals_ =  1 if _decimals_ is None else _decimals_

    # get the template values
    template = templates[_template_]

    # get the gene pool 
    gp = ghenv.Component.Params.Input[0].Sources[0]

    # modify the gene pool
    gp.Count = 24.0
    gp.Decimals = _decimals_
    gp.Name = "HB Day Schedule"
    gp.NickName = "DaySchedule"
    gp.Maximum = _up_bound_
    gp.Minimum = _low_bound_

    # set gene pool values
    for i in range(gp.Count):
        gp[i] = gp.Minimum + (template[i] * gp.Maximum)

    # register the change on the component
    gp.ExpireSolution(True)
else:
    print('Connect _gene_pool and a button to _run.')



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Create an EnergyPlus window construction. Material inputs can be either the
identifiers of materials within the library or a custom material made with any
of the EnergyPlus Material components.
_
Note that the _materials should be ordered from outermost (exterior) layer to the
innermost (interior) layer.
-

    Args:
        _name_: Text to set the name for the Construction and to be incorporated
            into a unique Construction identifier.
        _materials: List of materials in the construction (from exterior to interior).
            These materials can be either fully-detailed material objects built
            with the material components or text for a material identifier to be
            looked up in the window material library. Note that a native Grasshopper
            "Merge" component can be used to help order the materials correctly
            for the input here.

    Returns:
        constr: A window construction that can be assigned to Honeybee
            Apertures or ConstructionSets.
"""

ghenv.Component.Name = "HB Window Construction"
ghenv.Component.NickName = 'WindowConstr'
ghenv.Component.Message = '1.2.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = "1 :: Constructions"
ghenv.Component.AdditionalHelpFromDocStrings = "4"

try:  # import the core honeybee dependencies
    from honeybee.typing import clean_and_id_ep_string, clean_ep_string
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:  # import the honeybee-energy dependencies
    from honeybee_energy.construction.window import WindowConstruction
    from honeybee_energy.lib.materials import window_material_by_identifier
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:  # import ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    name = clean_and_id_ep_string('WindowConstruction') if _name_ is None else \
        clean_ep_string(_name_)

    material_objs = []
    for material in _materials:
        if isinstance(material, str):
            material = window_material_by_identifier(material)
        material_objs.append(material)

    constr = WindowConstruction(name, material_objs)
    if _name_ is not None:
        constr.display_name = _name_



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Create a standard opaque material, which can be plugged into the "HB Opaque
Construction" component.
-

    Args:
        _name_: Text to set the name for the material and to be incorporated into
            a unique material identifier.
        _thickness: Number for the thickness of the material layer [m].
        _conductivity: Number for the thermal conductivity of the material [W/m-K].
        _density: Number for the density of the material [kg/m3].
        _spec_heat: Number for the specific heat of the material [J/kg-K].
        _roughness_: Text describing the relative roughness of the material.
            Must be one of the following: 'VeryRough', 'Rough', 'MediumRough',
            'MediumSmooth', 'Smooth', 'VerySmooth'. (Default: 'MediumRough').
        _therm_absp_: A number between 0 and 1 for the fraction of incident long
            wavelength radiation that is absorbed by the material. (Default: 0.9).
        _sol_absp_: A number between 0 and 1 for the fraction of incident solar
            radiation absorbed by the material. (Default: 0.7).
        _vis_absp_: A number between 0 and 1 for the fraction of incident
            visible wavelength radiation absorbed by the material.
            Default value is the same as the _sol_absp_.

    Returns:
        mat: A standard opaque material that can be assigned to a Honeybee
            Opaque construction.
"""

ghenv.Component.Name = "HB Opaque Material"
ghenv.Component.NickName = 'OpaqueMat'
ghenv.Component.Message = '1.2.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = "1 :: Constructions"
ghenv.Component.AdditionalHelpFromDocStrings = "5"

try:  # import the core honeybee dependencies
    from honeybee.typing import clean_and_id_ep_string, clean_ep_string
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:  # import the honeybee-energy dependencies
    from honeybee_energy.material.opaque import EnergyMaterial
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:  # import ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # set the default material properties
    _roughness_ = 'MediumRough' if _roughness_ is None else _roughness_
    _therm_absp_ = 0.9 if _therm_absp_ is None else _therm_absp_
    _sol_absp_ = 0.7 if _sol_absp_ is None else _sol_absp_
    name = clean_and_id_ep_string('OpaqueMaterial') if _name_ is None else \
        clean_ep_string(_name_)

    # create the material
    mat = EnergyMaterial(
        name, _thickness, _conductivity, _density, _spec_heat, _roughness_,
        _therm_absp_, _sol_absp_, _vis_absp_)
    if _name_ is not None:
        mat.display_name = _name_



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Apply WindowConstruction or WindowConstructionShade to Honeybee Apertures or
glass Doors. Alternatively, it can assign WindowConstructions to the child
apertures of input Faces or the apertures within Room walls.
_
This component supports the assigning of different constructions based on cardinal
orientation, provided that a list of WindowConstructions are input to the _constr. 
-

    Args:
        _hb_objs: Honeybee Apertures, Faces, Doors or Rooms to which the input
            _constr should be assigned. For the case of a Honeybee Room, the
            construction will only be applied to the apertures in the the
            Room's outdoor walls. Note that, if you need to assign a construction
            to all the skylights, glass doors, etc. of a Room, the best practice
            is to create a ConstructionSet and assing that to the Room.
        _constr: A Honeybee WindowConstruction or WindowConstructionShade to be
            applied to the input _hb_objs. This can also be text for a construction
            to be looked up in the window construction library. If an array of
            text or construction objects are input here, different constructions
            will be assigned based on cardinal direction, starting with north
            and moving clockwise.
    
    Returns:
        hb_objs: The input honeybee objects with their constructions edited.
"""

ghenv.Component.Name = "HB Apply Window Construction"
ghenv.Component.NickName = 'ApplyWindowConstr'
ghenv.Component.Message = '1.2.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '1 :: Constructions'
ghenv.Component.AdditionalHelpFromDocStrings = '3'


try:  # import the honeybee-energy extension
    from honeybee_energy.lib.constructions import window_construction_by_identifier
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:  # import the core honeybee dependencies
    from honeybee.boundarycondition import Outdoors
    from honeybee.facetype import Wall
    from honeybee.room import Room
    from honeybee.face import Face
    from honeybee.aperture import Aperture
    from honeybee.door import Door
    from honeybee.orientation import angles_from_num_orient, face_orient_index
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:  # import the ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


def is_exterior_wall(face):
    """Check whether a given Face is an exterior Wall."""
    return isinstance(face.boundary_condition, Outdoors) and \
        isinstance(face.type, Wall)


if all_required_inputs(ghenv.Component):
    # duplicate the initial objects
    hb_objs = [obj.duplicate() for obj in _hb_objs]
    
    # process the input constructions
    for i, constr in enumerate(_constr):
        if isinstance(constr, str):
            _constr[i] = window_construction_by_identifier(constr)
    
    # error message for unrecognized object
    error_msg = 'Input _hb_objs must be a Room, Face, Aperture, or Door. Not {}.'
    
    # assign the constructions
    if len(_constr) == 1:  # assign indiscriminately, even if it's a horizontal object
        for obj in hb_objs:
            if isinstance(obj, (Aperture, Door)):
                obj.properties.energy.construction = _constr[0]
            elif isinstance(obj, Face):
                for ap in obj.apertures:
                    ap.properties.energy.construction = _constr[0]
            elif isinstance(obj, Room):
                for face in obj.faces:
                    if is_exterior_wall(face):
                        for ap in face.apertures:
                            ap.properties.energy.construction = _constr[0]
            else:
                raise TypeError(error_msg.format(type(obj)))
    else:  # assign constructions only to non-horizontal objects based on cardinal direction
        angles = angles_from_num_orient(len(_constr))
        for obj in hb_objs:
            if isinstance(obj, (Aperture, Door)):
                orient_i = face_orient_index(obj, angles)
                if orient_i is not None:
                    obj.properties.energy.construction = _constr[orient_i]
            elif isinstance(obj, Face):
                orient_i = face_orient_index(obj, angles)
                if orient_i is not None:
                    for ap in obj.apertures:
                        ap.properties.energy.construction = _constr[orient_i]
            elif isinstance(obj, Room):
                 for face in obj.faces:
                     if is_exterior_wall(face):
                         orient_i = face_orient_index(face, angles)
                         if orient_i is not None:
                            for ap in face.apertures:
                                ap.properties.energy.construction = _constr[orient_i]
            else:
                raise TypeError(error_msg.format(type(obj)))




# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Parse any time series data from an energy simulation SQL result file.

-
    Args:
        _sql: The file path of the SQL result file that has been generated from
            an energy simulation.
        _output_names: A list of EnergyPlus output names as strings (eg.
            'Surface Window System Solar Transmittance'. These data corresponding
            to these outputs will be returned from this component.
    
    Returns:
        results: DataCollections for the output_names.
"""

ghenv.Component.Name = 'HB Read Custom Result'
ghenv.Component.NickName = 'CustomResult'
ghenv.Component.Message = '1.2.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '6 :: Result'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

import os
import subprocess
import json

try:
    from ladybug.datacollection import HourlyContinuousCollection, \
        MonthlyCollection, DailyCollection
    from ladybug.sql import SQLiteResult
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from honeybee.config import folders
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))

def serialize_data(data_dicts):
    """Reserialize a list of collection dictionaries."""
    if len(data_dicts) == 0:
        return []
    elif data_dicts[0]['type'] == 'HourlyContinuousCollection':
        return [HourlyContinuousCollection.from_dict(data) for data in data_dicts]
    elif data_dicts[0]['type'] == 'MonthlyCollection':
        return [MonthlyCollection.from_dict(data) for data in data_dicts]
    elif data_dicts[0]['type'] == 'DailyCollection':
        return [DailyCollection.from_dict(data) for data in data_dicts]


if all_required_inputs(ghenv.Component):
    if os.name == 'nt':  # we are on windows; use IronPython like usual
        sql_obj = SQLiteResult(_sql)  # create the SQL result parsing object
        results = sql_obj.data_collections_by_output_name(_output_names)

    else:  # we are on Mac; sqlite3 module doesn't work in Mac IronPython
        # Execute the honybee CLI to obtain the results via CPython
        cmds = [folders.python_exe_path, '-m', 'honeybee_energy', 'result',
                'data-by-outputs', _sql, _output_names]
        process = subprocess.Popen(cmds, stdout=subprocess.PIPE)
        stdout = process.communicate()
        data_dicts = json.loads(stdout[0])
        results = serialize_data(data_dicts[0])



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Turn Honeybee Rooms into Plenums with no internal loads.
_
This includes removing all people, lighting, equipment, hot water, and mechanical
ventilation. By default, the heating/cooling system and setpoints will also be
removed but they can optionally be kept. Infiltration is kept by default but
can optionally be removed as well.
_
This is useful to appropriately assign properties for closets, underfloor spaces,
and drop ceilings.
-

    Args:
        _rooms: Honeybee Rooms to be converted into plenums.
        conditioned_: Boolean to indicate whether the plenum is conditioned with a
            heating/cooling system. If True, the setpoints of the Room will also
            be kept in addition to the heating/cooling system (Default: False).
        remove_infilt_: Boolean to indicate whether infiltration should be removed
            from the Rooms. (Default: False).

    Returns:
        rooms: Rooms that have had their interinal loads removed to reflect a
            plenum space.
"""

ghenv.Component.Name = 'HB Plenum'
ghenv.Component.NickName = 'Plenum'
ghenv.Component.Message = '1.2.1'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '0 :: Basic Properties'
ghenv.Component.AdditionalHelpFromDocStrings = '5'


try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    rooms = [room.duplicate() for room in _rooms]  # duplicate to avoid editing input
    for room in rooms:
        room.properties.energy.make_plenum(conditioned_, remove_infilt_)



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Parse all of the common Room-level comfort-related results from an SQL result
file that has been generated from an energy simulation.

-
    Args:
        _sql: The file path of the SQL result file that has been generated from
            an energy simulation.

    Returns:
        face_indoor_temp: DataCollections for the indoor surface temperature of
            each surface (C).
        face_outdoor_temp: DataCollections for the outdoor surface temperature
            of each surface (C).
        face_energy_flow: DataCollections for the heat loss (negative) or heat
            gain (positive) through each building surfaces (kWh).
"""

ghenv.Component.Name = 'HB Read Face Result'
ghenv.Component.NickName = 'FaceResult'
ghenv.Component.Message = '1.2.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '6 :: Result'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

import os
import subprocess
import json

try:
    from ladybug.datacollection import HourlyContinuousCollection, \
        MonthlyCollection, DailyCollection
    from ladybug.sql import SQLiteResult
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from honeybee.config import folders
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


def subtract_loss_from_gain(gain_load, loss_load):
    """Create a single DataCollection from gains and losses."""
    total_loads = []
    for gain, loss in zip(gain_load, loss_load):
        total_load = gain - loss
        total_load.header.metadata['type'] = \
            total_load.header.metadata['type'].replace('Gain ', '')
        total_loads.append(total_load)
    return total_loads


def serialize_data(data_dicts):
    """Reserialize a list of collection dictionaries."""
    if len(data_dicts) == 0:
        return []
    elif data_dicts[0]['type'] == 'HourlyContinuousCollection':
        return [HourlyContinuousCollection.from_dict(data) for data in data_dicts]
    elif data_dicts[0]['type'] == 'MonthlyCollection':
        return [MonthlyCollection.from_dict(data) for data in data_dicts]
    elif data_dicts[0]['type'] == 'DailyCollection':
        return [DailyCollection.from_dict(data) for data in data_dicts]


# List of all the output strings that will be requested
face_indoor_temp_output = 'Surface Inside Face Temperature'
face_outdoor_temp_output = 'Surface Outside Face Temperature'
opaque_energy_flow_output = 'Surface Average Face Conduction Heat Transfer Energy'
window_loss_output = 'Surface Window Heat Loss Energy'
window_gain_output = 'Surface Window Heat Gain Energy'
all_output = [face_indoor_temp_output, face_outdoor_temp_output,
              opaque_energy_flow_output, window_loss_output, window_gain_output]


if all_required_inputs(ghenv.Component):
    if os.name == 'nt':  # we are on windows; use IronPython like usual
        sql_obj = SQLiteResult(_sql)  # create the SQL result parsing object
        # get all of the results
        face_indoor_temp = sql_obj.data_collections_by_output_name(face_indoor_temp_output)
        face_outdoor_temp = sql_obj.data_collections_by_output_name(face_outdoor_temp_output)
        opaque_energy_flow = sql_obj.data_collections_by_output_name(opaque_energy_flow_output)
        window_loss = sql_obj.data_collections_by_output_name(window_loss_output)
        window_gain = sql_obj.data_collections_by_output_name(window_gain_output)

    else:  # we are on Mac; sqlite3 module doesn't work in Mac IronPython
        # Execute the honybee CLI to obtain the results via CPython
        cmds = [folders.python_exe_path, '-m', 'honeybee_energy', 'result',
                'data-by-outputs', _sql] + all_output
        process = subprocess.Popen(cmds, stdout=subprocess.PIPE)
        stdout = process.communicate()
        data_coll_dicts = json.loads(stdout[0])
        # get all of the results
        face_indoor_temp = serialize_data(data_coll_dicts[0])
        face_outdoor_temp = serialize_data(data_coll_dicts[1])
        opaque_energy_flow = serialize_data(data_coll_dicts[2])
        window_loss = serialize_data(data_coll_dicts[3])
        window_gain = serialize_data(data_coll_dicts[4])

    # do arithmetic with any of the gain/loss data collections
    window_energy_flow = []
    if len(window_gain) == len(window_loss):
        window_energy_flow = subtract_loss_from_gain(window_gain, window_loss)
    face_energy_flow = opaque_energy_flow + window_energy_flow



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Run Honeybee Rooms through a quick energy simulation to obtain an estimate of
room-level peak cooling and heating on summer and winter design days.
-

    Args:
        _rooms: A list of Honeybee Rooms for which peak loads will be computed.
        shades_: An optional list of Honeybee Shades that can block the sun to
            the input _rooms.
        _ddy_file: Path to a .ddy file on your system as a text string, which contains
            design day conditions for the peak load analysis. This can also
            be the path to an .epw file, in which case design days will be
            determined by statitically analysing the annual data to approximate
            0.4% and 99.6% design conditions. Note that .ddy files can also be
            created using the "DF Construct Design Day" and "DF Write DDY"
            components.
            _
            When constructing custom design days, it is recommended that the .ddy
            file contain only one summer and one winter design day. If mutliple
            summer or winter design days are found, they will be filtered according
            to their name in order to identify the 0.4% and 99.6% design conditions
            for the sensible (dry bulb temperature) design days.
        _timestep_: An integer for the number of timesteps per hour at which the energy
            simulation will be run and results reported. It is recommended that
            this be at least 6 but it can be increased to better capture
            the minute in which peak cooling occurs. (Default: 6).
            The following values are acceptable:
            (1, 2, 3, 4, 5, 6, 10, 12, 15, 20, 30, 60)
        run_bal_: Set to True to have the load balance computed after the
            simulation is run. This ensures that data collections for various
            terms of the load balance are output from the "balance".
            This can help explain why the loads are what they are but can
            also increase the component run time. (Default: False).
        _run: Set to "True" to run the simulation to obtain annual loads. This can
            also be the integer 2 to run the simulation while being able to see
            the simulation process (with a batch window).

    Returns:
        report: A report of the energy simulation run.
        peak_cool: A list of numbers that align with the input _rooms and correspond
            to the peak cooling of each room on the summer design day in Watts.
            Note that, for multi-room simulations, the individual room peaks may
            not be coincident, meaning that summing these values together won't
            give a correct sense of the size of cetral cooling equipment serving
            multiple rooms. For such equipment, the max of the cooling data
            collection should be used.
        peak_heat: A list of numbers that align with the input _rooms and correspond
            to the peak heating of each room on the winter design day in Watts.
            Note that, for multi-room simulations, the individual room peaks may
            not be coincident, meaning that summing these values together won't
            give a correct sense of the size of cetral heating equipment serving
            multiple rooms. For such equipment, the max of the heating data
            collection should be used.
        cooling: A Data Collection indicating the combined cooling demand of the rooms
            at each simulation timestep of the summer design day. This can be
            plugged into the "LB Monthly Chart" component to visualize the demand
            or it can be deconstructed with the "LB Deconstruct Data" component
            for analysis.
        heating: A Data Collection indicating the combined heating demand of the rooms
            at each simulation timestepof the winter design day. This can be
            plugged into the "LB Monthly Chart" component to visualize the demand
            or it can be deconstructed with the "LB Deconstruct Data" component
            for analysis.
        cool_bal: A list of data collections for the various terms of the sensible load
            balance that contribute to peak cooling on the summer design day. These
            can be plugged into the "LB Monthly Chart" component (with stack_ set
            to True) to visualize the terms contributing to the peak. Will be
            None unless run_bal_ is set to True.
        heat_bal: A list of data collections for the various terms of the sensible load
            balance that contribute to peak heating on the summer design day. These
            can be plugged into the "LB Monthly Chart" component (with stack_ set
            to True) to visualize the terms contributing to the peak. Will be
            None unless run_bal_ is set to True.
"""

ghenv.Component.Name = 'HB Peak Loads'
ghenv.Component.NickName = 'PeakLoads'
ghenv.Component.Message = '1.2.1'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '5 :: Simulate'
ghenv.Component.AdditionalHelpFromDocStrings = '2'

import os
import subprocess
import json

try:
    from ladybug.futil import write_to_file_by_name, nukedir
    from ladybug.ddy import DDY
    from ladybug.epw import EPW
    from ladybug.sql import SQLiteResult
    from ladybug.datacollection import HourlyContinuousCollection
    from ladybug.header import Header
    from ladybug.analysisperiod import AnalysisPeriod
    from ladybug.datatype.power import Power
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from honeybee.config import folders
    from honeybee.model import Model
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:
    from honeybee_energy.result.loadbalance import LoadBalance
    from honeybee_energy.simulation.parameter import SimulationParameter
    from honeybee_energy.run import run_idf
    from honeybee_energy.result.err import Err
    from honeybee_energy.result.zsz import ZSZ
    from honeybee_energy.writer import energyplus_idf_version
    from honeybee_energy.config import folders as energy_folders
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from lbt_recipes.version import check_energyplus_version
except ImportError as e:
    raise ImportError('\nFailed to import lbt_recipes:\n\t{}'.format(e))

try:
    from ladybug_rhino.config import tolerance, angle_tolerance
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


def check_for_filter_failure(des_days):
    """Raise a ValueError in the event that the design-dsy filtering process failed."""
    if len(des_days) == 0:
        raise ValueError(
            'Failed to filter the design days in the .ddy file to find the most '
            'appropriate one for sensible peak loads.\nTry connecting an .epw file '
            'instead.\n Or try creating your own .ddy file with a single summer '
            'and winter design day.'
        )


def check_and_filter_des_days(sim_par, des_days, day_type):
    """Filter design days to get the most appropriate one and assing it to sim_par."""
    if len(des_days) == 0:
        raise ValueError('No {}s were found in the connected .ddy file.'.format(day_type))
    elif len(des_days) == 1:  # just assign the one design day
        _sim_par_.sizing_parameter.add_design_day(des_days[0])
    else:  # find the most appropriate design day by percent
        key_wrd = '99.6%' if day_type == 'WinterDesignDay' else '.4%'
        des_days = [dday for dday in des_days if key_wrd in dday.name]
        check_for_filter_failure(des_days)
        if len(des_days) == 1:
            _sim_par_.sizing_parameter.add_design_day(des_days[0])
        else:  # find the most appropriate design day bu type
            key_wrd = ' DB' if day_type == 'WinterDesignDay' else ' DB=>MWB'
            des_days = [dday for dday in des_days if key_wrd in dday.name]
            check_for_filter_failure(des_days)
            if len(des_days) == 1:
                _sim_par_.sizing_parameter.add_design_day(des_days[0])
            else:
                check_for_filter_failure([])


def data_to_load(data_colls, data_type, analysis_period):
    """Convert data collections output by EnergyPlus to a single load collection.

    Args:
        data_colls: A list of monthly data collections for an energy term.
        data_type: Text for the data type of the collections (eg. "Cooling").
        analysis_period: AnalysisPeriod object describing the date and timestep
            of the design day.
    """
    if len(data_colls) != 0:
        total_vals = [sum(ts_vals) for ts_vals in zip(*data_colls)]
    else:  # just make a "filler" collection of 0 values
        total_vals = [0] * (24 * analysis_period.timestep)
    meta_dat = {'type': data_type}
    total_head = Header(Power(), 'W', analysis_period, meta_dat)
    return HourlyContinuousCollection(total_head, total_vals)


def serialize_data(data_dicts):
    """Reserialize a list of collection dictionaries."""
    return [HourlyContinuousCollection.from_dict(dat) for dat in data_dicts]


def filter_data_by_date(filt_date, data):
    """Filter a matrix of data collections by the date in their analysis period."""
    filtered_data = []
    for data_list in data:
        flit_list = []
        for d in data_list:
            a_per = d.header.analysis_period
            if a_per.st_month == filt_date.month and a_per.st_day == filt_date.day:
                flit_list.append(d.to_time_rate_of_change())
        filtered_data.append(flit_list)
    return filtered_data


def all_data_load_balance(rooms, data):
    """Get a LoadBalance object from a list of all relavant data collections."""
    return LoadBalance(
        rooms, lighting_data=data[0], electric_equip_data=data[1],
        gas_equip_data=c_data[2], service_hot_water_data=data[3],
        people_data=data[4], solar_data=data[5], infiltration_data=data[6],
        surface_flow_data=data[7], use_all_solar=True
    )


def reorder_balance(balance, order):
    """Reorder the terms of a load balance according to the desired names."""
    new_balance = []
    for term_name in order:
        for term in balance:
            if term.header.metadata['type'] == term_name:
                new_balance.append(term)
                break
    return new_balance


# List of the output strings that will be requested
opaque_energy_flow_output = 'Surface Average Face Conduction Heat Transfer Energy'
window_loss_output = 'Surface Window Heat Loss Energy'
window_gain_output = 'Surface Window Heat Gain Energy'
all_output = \
    [LoadBalance.LIGHTING, LoadBalance.ELECTRIC_EQUIP, LoadBalance.GAS_EQUIP,
     LoadBalance.HOT_WATER, LoadBalance.PEOPLE_GAIN, LoadBalance.SOLAR_GAIN,
     LoadBalance.INFIL_GAIN, LoadBalance.INFIL_LOSS, opaque_energy_flow_output,
     window_loss_output, window_gain_output]
term_order = \
    ['Solar', 'Window Conduction', 'Opaque Conduction', 'Infiltration', 'People',
     'Lighting', 'Electric Equipment', 'Gas Equipment', 'Service Hot Water']


if all_required_inputs(ghenv.Component) and _run:
    # check the presence of energyplus and check that the version is compatible
    check_energyplus_version()

    # create the Model from the _rooms and shades_
    _model = Model('Peak_Loads', _rooms, orphaned_shades=shades_,
                   tolerance=tolerance, angle_tolerance=angle_tolerance)

    # process the simulation folder name and the directory
    directory = os.path.join(folders.default_simulation_folder, _model.identifier)
    sch_directory = os.path.join(directory, 'schedules')
    nukedir(directory)  # delete any existing files in the directory

    # create simulation parameters for a design-day-optimized E+ sim
    _sim_par_ = SimulationParameter()
    _sim_par_.timestep = _timestep_ if _timestep_ is not None else 6
    _sim_par_.output.reporting_frequency = 'Timestep'
    _sim_par_.simulation_control.run_for_sizing_periods = True
    _sim_par_.simulation_control.run_for_run_periods = False
    if run_bal_:
        _sim_par_.output.add_zone_energy_use('Sensible')
        _sim_par_.output.add_gains_and_losses('Sensible')
        _sim_par_.output.add_surface_energy_flow()

    # load design days to the simulation parameters
    if _ddy_file.lower().endswith('.epw'):  # load design days from EPW
        epw_obj = EPW(_ddy_file)
        des_days = epw_obj.best_available_design_days()
        _sim_par_.sizing_parameter.design_days = des_days
        location = epw_obj.location
    else:  # load design days from DDY
        ddy_obj = DDY.from_ddy_file(_ddy_file)
        w_days = [day for day in ddy_obj.design_days if day.day_type == 'WinterDesignDay']
        s_days = [day for day in ddy_obj.design_days if day.day_type == 'SummerDesignDay']
        check_and_filter_des_days(_sim_par_, w_days, 'WinterDesignDay')
        check_and_filter_des_days(_sim_par_, s_days, 'SummerDesignDay')
        location = ddy_obj.location

    # get the dates of the heating and cooling design days
    h_dt = _sim_par_.sizing_parameter.design_days[0].sky_condition.date
    c_dt = _sim_par_.sizing_parameter.design_days[1].sky_condition.date
    tst = _sim_par_.timestep
    heat_ap = AnalysisPeriod(h_dt.month, h_dt.day, 0, h_dt.month, h_dt.day, 23, tst)
    cool_ap = AnalysisPeriod(c_dt.month, c_dt.day, 0, c_dt.month, c_dt.day, 23, tst)

    # create the strings for simulation paramters and model
    ver_str = energyplus_idf_version() if energy_folders.energyplus_version \
        is not None else energyplus_idf_version(compatibe_ep_version)
    model_str = _model.to.idf(_model, schedule_directory=sch_directory)
    idf_str = '\n\n'.join([ver_str, location.to_idf(), _sim_par_.to_idf(), model_str])

    # write the final string into an IDF
    idf = os.path.join(directory, 'in.idf')
    write_to_file_by_name(directory, 'in.idf', idf_str, True)

    # run the IDF through EnergyPlus
    silent = True if _run == 1 else False
    sql, zsz, rdd, html, err = run_idf(idf, silent=silent)
    if html is None and err is not None:  # something went wrong; parse the errors
        err_obj = Err(err)
        print(err_obj.file_contents)
        for error in err_obj.fatal_errors:
            raise Exception(error)

    # parse the result ZSZ and get the timestep data collections
    zsz_obj = ZSZ(zsz)
    cool_init = zsz_obj.cooling_load_data
    heat_init = zsz_obj.heating_load_data
    cooling = data_to_load(cool_init, 'Cooling', cool_ap)
    heating = data_to_load(heat_init, 'Heating', heat_ap)

    # parse the result sql and get the timestep data collections
    if os.name == 'nt':  # we are on windows; use IronPython like usual
        sql_obj = SQLiteResult(sql)
        peak_cool_dict = {zs.zone_name: zs.calculated_design_load
                          for zs in sql_obj.zone_cooling_sizes}
        peak_heat_dict = {zs.zone_name: zs.calculated_design_load
                          for zs in sql_obj.zone_heating_sizes}
    else:  # we are on Mac; sqlite3 module doesn't work in Mac IronPython
        # Execute the honybee CLI to obtain the results via CPython
        cmds = [folders.python_exe_path, '-m', 'honeybee_energy', 'result',
                'zone-sizes', sql]
        process = subprocess.Popen(cmds, stdout=subprocess.PIPE)
        stdout = process.communicate()
        peak_dicts = json.loads(stdout[0])
        peak_cool_dict = {zs['zone_name']: zs['calculated_design_load']
                          for zs in peak_dicts['cooling']}
        peak_heat_dict = {zs['zone_name']: zs['calculated_design_load']
                          for zs in peak_dicts['heating']}
    peak_cool, peak_heat = [], []
    for rm in _rooms:
        rm_id = rm.identifier.upper()
        try:
            peak_cool.append(peak_cool_dict[rm_id])
        except KeyError:
            peak_cool.append(0)
        try:
            peak_heat.append(peak_heat_dict[rm_id])
        except KeyError:
            peak_heat.append(0)

    # construct the load balance if requested
    if run_bal_:
        if os.name == 'nt':  # we are on windows; use IronPython like usual
            light = sql_obj.data_collections_by_output_name(LoadBalance.LIGHTING)
            ele_equip = sql_obj.data_collections_by_output_name(LoadBalance.ELECTRIC_EQUIP)
            gas_equip = sql_obj.data_collections_by_output_name(LoadBalance.GAS_EQUIP)
            hot_water = sql_obj.data_collections_by_output_name(LoadBalance.HOT_WATER)
            people = sql_obj.data_collections_by_output_name(LoadBalance.PEOPLE_GAIN)
            solar = sql_obj.data_collections_by_output_name(LoadBalance.SOLAR_GAIN)
            infil_gain = sql_obj.data_collections_by_output_name(LoadBalance.INFIL_GAIN)
            infil_loss = sql_obj.data_collections_by_output_name(LoadBalance.INFIL_LOSS)
            opaque_flow = sql_obj.data_collections_by_output_name(opaque_energy_flow_output)
            window_loss = sql_obj.data_collections_by_output_name(window_loss_output)
            window_gain = sql_obj.data_collections_by_output_name(window_gain_output)
        else:  # we are on Mac; sqlite3 module doesn't work in Mac IronPython
            # Execute the honybee CLI to obtain the results via CPython
            cmds = [folders.python_exe_path, '-m', 'honeybee_energy', 'result',
                    'data-by-outputs', sql]
            for outp in all_output:
                out_str = json.dumps(outp) if isinstance(outp, tuple) else '["{}"]'.format(outp)
                cmds.append(out_str)
            process = subprocess.Popen(cmds, stdout=subprocess.PIPE)
            stdout = process.communicate()
            data_coll_dicts = json.loads(stdout[0])
            light = serialize_data(data_coll_dicts[0])
            ele_equip = serialize_data(data_coll_dicts[1])
            gas_equip = serialize_data(data_coll_dicts[2])
            hot_water = serialize_data(data_coll_dicts[3])
            people = serialize_data(data_coll_dicts[4])
            solar = serialize_data(data_coll_dicts[5])
            infil_gain = serialize_data(data_coll_dicts[6])
            infil_loss = serialize_data(data_coll_dicts[7])
            opaque_flow = serialize_data(data_coll_dicts[8])
            window_loss = serialize_data(data_coll_dicts[9])
            window_gain = serialize_data(data_coll_dicts[10])
            for dat in opaque_flow + window_loss + window_gain:
                dat.header.metadata['Surface'] = dat.header.metadata['Zone']

        infil = LoadBalance.subtract_loss_from_gain(infil_gain, infil_loss)
        window_flow = []
        window_flow = LoadBalance.subtract_loss_from_gain(window_gain, window_loss)
        face_flow = opaque_flow + window_flow
        all_data = [light, ele_equip, gas_equip, hot_water, people, solar, infil, face_flow]

        # construct the cooling design day balance
        c_data = filter_data_by_date(c_dt, all_data)
        load_bal = all_data_load_balance(_rooms, c_data)
        cool_bal = reorder_balance(load_bal.load_balance_terms(), term_order)

        # construct the heating design day balance
        h_data = filter_data_by_date(h_dt, all_data)
        load_bal = all_data_load_balance(_rooms, h_data)
        heat_bal = reorder_balance(load_bal.load_balance_terms(), term_order)



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Create an Equipment object that can be used to specify equipment usage in a ProgramType.
-

    Args:
        _name_: Text to set the name for the Equipment and to be incorporated
            into a unique Equipment identifier. If None, a unique name will
            be generated.
         _watts_per_area: A numerical value for the equipment power density in
            Watts per square meter of floor area.
        _schedule: A fractional schedule for the use of equipment over the course of
            the year. The fractional values will get multiplied by the
            _watts_per_area to yield a complete equipment profile.
        radiant_fract_: A number between 0 and 1 for the fraction of the total
            equipment load given off as long wave radiant heat. (Default: 0).
        latent_fract_: A number between 0 and 1 for the fraction of the total
            equipment load that is latent (as opposed to sensible). (Default: 0).
        lost_fract_: A number between 0 and 1 for the fraction of the total
            equipment load that is lost outside of the zone and the HVAC system.
            Typically, this is used to represent heat that is exhausted directly
            out of a zone (as you would for a stove). (Default: 0).
        gas_: Set to "True" to have the output Equipment object be for GasEquipment
            (as opposed to ElectricEquipment).

    Returns:
        equip: An Equipment object that can be used to specify equipment usage in
            a ProgramType.
"""

ghenv.Component.Name = 'HB Equipment'
ghenv.Component.NickName = 'Equipment'
ghenv.Component.Message = '1.2.1'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '3 :: Loads'
ghenv.Component.AdditionalHelpFromDocStrings = '3'

try:  # import the core honeybee dependencies
    from honeybee.typing import clean_and_id_ep_string, clean_ep_string
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:
    from honeybee_energy.load.equipment import ElectricEquipment, GasEquipment
    from honeybee_energy.lib.schedules import schedule_by_identifier
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # make a default Equipment name if none is provided
    name = clean_and_id_ep_string('Equipment') if _name_ is None else \
        clean_ep_string(_name_)

    # get the schedule
    if isinstance(_schedule, str):
        _schedule = schedule_by_identifier(_schedule)

    # get default radiant, latent, and lost fractions
    radiant_fract_ = radiant_fract_ if radiant_fract_ is not None else 0.0
    latent_fract_ = latent_fract_ if latent_fract_ is not None else 0.0
    lost_fract_ = lost_fract_ if lost_fract_ is not None else 0.0

    # create the Equipment object
    if gas_:
        equip = GasEquipment(name, _watts_per_area, _schedule,
                             radiant_fract_, latent_fract_, lost_fract_)
    else:
        equip = ElectricEquipment(name, _watts_per_area, _schedule,
                                  radiant_fract_, latent_fract_, lost_fract_)
    if _name_ is not None:
        equip.display_name = _name_


# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Set up a Honeybee Model to use the EnergyPlus Airflow Network (AFN) for all airflow
in the energy simulation.
_
Compared to the default single-zone methods that Honeybee uses for infiltration
and ventilation, the AFN represents air flow in a manner that is truer to the fluid
dynamic behavior of real buildings. In particular, the AFN more accurately models
the flow of air from one zone to another, accounting for the pressure changes
induced by wind and air density differences. However, using the AFN means that
the simulation will take considerably longer to run compared to the single zone
option and the difference in simuation resuts is only likely to be significant
when the Model contains operable windows or the building is extremely leaky.
_
Passing a Honeybee Model through this component before energy simulation will
result in the following changes to the EnergyPlus IDF:
_
1. All ZoneInfiltration objects will be excluded and, instead, infitration will
be modeled with AFN Crack objects assigned to each opaque Face.
_
2. For all AirBoundary Faces within the Model, ZoneMixing objects will be excluded
and, instead, the air boundary will be modeled with AFN Crack objects that have
very large pressure coefficients derived from the orifice equation and the area
of the air wall.
_
3. For all operable Apertures, ZoneVentilation:WindandStackOpenArea objects will
be excluded and, instead, these opearable apertures will be modeled with AFN
SimpleOpening objects.
_
4. For each Room with a VentilationControl object to specify setpoints at which
the windows open, an Energy Management System (EMS) program will be written to
dictate when the operable Apertures of the Room open.
-

    Args:
        _model: A Honeybee Model for which the Airflow network will be set up.
            This Model should have everything assigned to it that is needed
            for simulation, including solved adjacencies and relevant window-
            opening properties.
        leakage_template_: Text identifying the leakiness of the Model, which is used
            to generate AFNCrack objects that represent infiltration for
            each of the Model's surfaces (Face, Aperture, and Door).
            _
            Choose from the following.
            * Excellent
            * Medium
            * VeryPoor
            _
            These three text values correspond to DesignBuilder's Cracks
            Templates, which provide typical crack flow coefficients and
            exponents for different envelope tightness classifications.
            _
            If None, the exterior airflow leakage parameters will be derived
            from the room infiltration rate specified in the room's energy
            properties, which are in units of m3/s per m2 of facade. This
            derivation from the Room's infiltration will compute air leakage
            parameters for exterior cracks that produce a total air flow rate
            equivalent to the room's infiltration rate at an envelope pressure
            difference of 4 Pa. This default derivation is not as complete of a
            representation of building ariflow dynamics as the DesignBuilder
            Crack Templates are. However, since the airflow leakae parameters are
            derived from values in m3/s-m2 of infiltration, they are easier to
            relate to the results of infiltration blower-door tests, which
            typically express infiltration rates in these units.
        _delta_pressure_: The air pressure difference across the building envelope in
            Pascals, which is used to calculate infiltration crack flow
            coefficients when no leakage tempate is specified. The resulting
            average simulated air pressure difference will roughly equal this
            delta pressure times the nth root of the ratio between the simulated
            and target room infiltration rates. (Default: 4).
        _ref_pressure_: The reference barometric pressure measurement in Pascals
            under which the surface crack data were obtained. (Default: 101325).
        _high_rise_: Booling indicating whether the Model is LowRise or HighRise.
            This parameter is used to estimate building-wide wind pressure
            coefficients for the AFN by approximating the building geometry
            as an extruded rectangle. LowRise corresponds to a building where
            the height is less then three times the width AND length of the
            footprint. HighRise corresponds to a building where height is more
            than three times the width OR length of the footprint. If None,
            this property will be auto-calculated from Room geometry of the
            Model. This default assumption may not be appropriate if the Model
            represents only a portion of a larger Building.
        _long_axis_: A number between 0 and 180 for the clockwise angle difference
            in degrees that the long axis of the building is from true North.
            This parameter is used to estimate building-wide wind pressure
            coefficients for the AFN by approximating the building geometry
            as an extruded rectangle. 0 indicates a North-South long axis while
            90 indicates an East-West long axis. If None, this property will be
            auto-calculated from Room geometry of the Model (either 0 or 90).
            This default assumption may not be appropriate if the Model
            represents only a portion of a larger Building.
        _aspect_ratio_: A number between 0 and 1 for the aspect ratio of the building's
            footprint, defined as the ratio of length of the short axis divided
            by the length of the long axis. This parameter is used to estimate
            building-wide wind pressure coefficients for the AFN by approximating
            the building geometry as an extruded rectangle If None, this
            property will be auto-calculated from Room geometry of the Model
            and the _long_axis_ above. This default assumption may not be
            appropriate if the Model represents only a portion of a larger
            building.

    Returns:
        model: The input Honeybee Model for which the Airflow network has
            been set up.
"""

ghenv.Component.Name = 'HB Airflow Newtwork'
ghenv.Component.NickName = 'AFN'
ghenv.Component.Message = '1.2.1'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '3 :: Loads'
ghenv.Component.AdditionalHelpFromDocStrings = '4'

try:
    from honeybee_energy.ventcool.afn import generate
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs, give_warning, \
        longest_list
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))

leakage_tempaltes = {
    'excellent': 'Excellent',
    'medium': 'Medium',
    'verypoor': 'VeryPoor'
}


if all_required_inputs(ghenv.Component):
    # duplicate the input Model to avoid editing it
    model = _model.duplicate()

    # set default properties for the leakage if they are not input
    try:
        leakage = leakage_tempaltes[leakage_template_.lower()] \
            if leakage_template_ is not None else 'Medium'
    except KeyError:
        raise TypeError('leakage_template_ "{}" is not recognized. Choose from: '
                        'Excellent, Medium VeryPoor'.format(leakage_template_))
    use_room_infiltration = True if leakage_template_ is None else False
    pressure = _ref_pressure_ if _ref_pressure_ is not None else 101325
    delta_pressure = _delta_pressure_ if _delta_pressure_ is not None else 4

    # generate the AFN leakage for all of the surfaces of the Model
    generate(model.rooms, leakage, use_room_infiltration, pressure, delta_pressure)

    # set up the Model-wide VentilationSimulationParameters for the AFN
    vent_sim_par = model.properties.energy.ventilation_simulation_control
    vent_sim_par.vent_control_type = 'MultiZoneWithoutDistribution'
    if _long_axis_ is not None:  # assing this first so it's in the autocalculation
        vent_sim_par.long_axis_angle = _long_axis_
    model.properties.energy.autocalculate_ventilation_simulation_control()

    # set the properties used to approximate wind pressure coefficients
    if _high_rise_ is not None:
        vent_sim_par.building_type = 'HighRise' if _high_rise_ else 'LowRise'
    if _aspect_ratio_ is not None:
        vent_sim_par.aspect_ratio = _aspect_ratio_
        vent_sim_par.long_axis_angle = _long_axis_
    report = model.properties.energy.ventilation_simulation_control



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Run an IDF file through EnergyPlus.

-
    Args:
        _idf: Path to an IDF file as a text string. This can also be a list of
            IDF files.
        _epw_ile: Path to an .epw file as a text string.
        parallel_: Set to "True" to run execute simulations of multiple IDF files
            in parallel, which can greatly increase the speed of calculation but
            may not be desired when other processes are running. If False, all
            EnergyPlus simulations will be be run on a single core. Default: False.
        _run: Set to "True" to run the IDF through EnergyPlus.
            _
            This input can also be the integer "2", which will run the whole
            simulation silently (without any batch windows).

    Returns:
        report: Check here to see a report of the EnergyPlus run.
        sql: The file path of the SQL result file that has been generated on your
            machine.
        zsz: Path to a .csv file containing detailed zone load information recorded
            over the course of the design days.
        rdd: The file path of the Result Data Dictionary (.rdd) file that is
            generated after running the file through EnergyPlus.  This file
            contains all possible outputs that can be requested from the EnergyPlus
            model.  Use the Read Result Dictionary component to see what outputs
            can be requested.
        html: The HTML file path of the Summary Reports. Note that this will be None
            unless the input _sim_par_ denotes that an HTML report is requested and
            _run is set to True.
"""

ghenv.Component.Name = 'HB Run IDF'
ghenv.Component.NickName = 'RunIDF'
ghenv.Component.Message = '1.2.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '5 :: Simulate'
ghenv.Component.AdditionalHelpFromDocStrings = '5'

import os
import shutil
import System.Threading.Tasks as tasks

try:
    from ladybug.futil import preparedir
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from honeybee_energy.run import run_idf
    from honeybee_energy.result.err import Err
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs, give_warning
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


def run_idf_and_report_errors(i):
    """Run an IDF file through EnergyPlus and report errors/warnings on this component."""
    idf = idfs[i]
    sql_i, zsz_i, rdd_i, html_i, err_i = run_idf(idf, _epw_file, silent=silent)

    # report any errors on this component
    if err_i is not None:
        err_obj = Err(err_i)
        err_objs.append(err_obj)
        for warn in err_obj.severe_errors:
            give_warning(ghenv.Component, warn)
        for error in err_obj.fatal_errors:
            print err_obj.file_contents  # print before raising the error
            raise Exception(error)

    # append everything to the global lists
    sql.append(sql_i)
    zsz.append(zsz_i)
    rdd.append(rdd_i)
    html.append(html_i)
    err.append(err_i)


if all_required_inputs(ghenv.Component) and _run:
    # global lists of outputs to be filled
    sql, zsz, rdd, html, err, err_objs = [], [], [], [], [], []

    # copy the IDFs into a sub-directory if they are not already labeled as in.idf
    idfs = []
    for idf_file_path in _idf:
        idf_dir, idf_file_name = os.path.split(idf_file_path)
        if idf_file_name != 'in.idf':  # copy the IDF file into a sub-directory
            sub_dir = os.path.join(idf_dir, 'run')
            target = os.path.join(sub_dir, 'in.idf')
            preparedir(sub_dir)
            shutil.copy(idf_file_path, target)
            idfs.append(target)
        else:
            idfs.append(idf_file_path)

    # run the IDF files through E+
    silent = True if _run == 2 else False
    if parallel_:
        tasks.Parallel.ForEach(range(len(idfs)), run_idf_and_report_errors)
    else:
        for i in range(len(idfs)):
            run_idf_and_report_errors(i)

    # print out error report if it's only one
    # otherwise it's too much data to be read-able
    if len(err_objs) == 1:
        print(err_objs[0].file_contents)



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Apply a customized IdealAirSystem to Honeybee Rooms.
-

    Args:
        _rooms: Honeybee Rooms to which the input ideal air properties will
            be assigned. This can also be a Honeybee Model for which all
            conditioned Rooms will be assigned the ideal air system.
        _economizer_: Text to indicate the type of air-side economizer used on
            the ideal air system. Economizers will mix in a greater amount of
            outdoor air to cool the zone (rather than running the cooling system)
            when the zone needs cooling and the outdoor air is cooler than the zone.
            Choose from the options below. Default: DifferentialDryBulb.
                * NoEconomizer
                * DifferentialDryBulb
                * DifferentialEnthalpy
        dcv_: Boolean to note whether demand controlled ventilation should be
            used on the system, which will vary the amount of ventilation air
            according to the occupancy schedule of the zone. Default: False.
        sensible_hr_: A number between 0 and 1 for the effectiveness of sensible
            heat recovery within the system. Default: 0.
        latent_hr_: A number between 0 and 1 for the effectiveness of latent heat
            recovery within the system. Default: 0.
        _heat_temp_: A number for the maximum heating supply air temperature
            [C]. Default: 50, which is typical for many air-based HVAC systems.
        _cool_temp_: A number for the minimum cooling supply air temperature
            [C]. Default: 13, which is typical for many air-based HVAC systems.
        _heat_limit_: A number for the maximum heating capacity in Watts. This
            can also be the text 'autosize' to indicate that the capacity should
            be determined during the EnergyPlus sizing calculation. This can also
            be the text 'NoLimit' to indicate no upper limit to the heating
            capacity. Default: 'autosize'.
        _cool_limit_: A number for the maximum cooling capacity in Watts. This
            can also be the text 'autosize' to indicate that the capacity should
            be determined during the EnergyPlus sizing calculation. This can also
            be the text 'NoLimit' to indicate no upper limit to the cooling
            capacity. Default: 'autosize'.
    
    Returns:
        rooms: The input Rooms with their Ideal Air Systems edited.
"""

ghenv.Component.Name = "HB IdealAir"
ghenv.Component.NickName = 'IdealAir'
ghenv.Component.Message = '1.2.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '4 :: HVAC'
ghenv.Component.AdditionalHelpFromDocStrings = '1'


try:  # import the honeybee extension
    from honeybee.altnumber import autosize, no_limit
    from honeybee.model import Model
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:  # import the honeybee-energy extension
    from honeybee_energy.hvac.idealair import IdealAirSystem
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))

# dictionary to get alterante number types
alt_numbers = {
    'nolimit': no_limit,
    'NoLimit': no_limit,
    'autosize': autosize,
    'Autosize': autosize,
    None: autosize
    }

if all_required_inputs(ghenv.Component):
    # extract any rooms from input Models and duplicate the rooms
    rooms = []
    for hb_obj in _rooms:
        if isinstance(hb_obj, Model):
            rooms.extend([room.duplicate() for room in hb_obj.rooms])
        else:
            rooms.append(hb_obj.duplicate())

    for room in rooms:
        if room.properties.energy.is_conditioned:
            # check to be sure the assigned HVAC system is an IdealAirSystem
            if not isinstance(room.properties.energy.hvac, IdealAirSystem):
                room.properties.energy.add_default_ideal_air()
            
            # create the customized ideal air system
            new_ideal_air = room.properties.energy.hvac.duplicate()
            if _economizer_ is not None:
                new_ideal_air.economizer_type = _economizer_
            if dcv_ is not None:
                new_ideal_air.demand_controlled_ventilation = dcv_
            if sensible_hr_ is not None:
                new_ideal_air.sensible_heat_recovery = sensible_hr_
            if latent_hr_ is not None:
                new_ideal_air.latent_heat_recovery = latent_hr_
            if _heat_temp_ is not None:
                new_ideal_air.heating_air_temperature = _heat_temp_
            if _cool_temp_ is not None:
                new_ideal_air.cooling_air_temperature = _cool_temp_
            try:
                new_ideal_air.heating_limit = alt_numbers[_heat_limit_]
            except KeyError:
                new_ideal_air.heating_limit = _heat_limit_
            try:
                new_ideal_air.cooling_limit = alt_numbers[_cool_limit_]
            except KeyError:
                new_ideal_air.cooling_limit = _cool_limit_
            
            # assign the HVAC to the Room
            room.properties.energy.hvac = new_ideal_air



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Create a ProgramType object possessing all schedules and loads defining the
usage of a space.
_
ProgramTypes can be assigned to Honeybee Rooms to specify all default schedules
and loads on the Room.
-

    Args:
        _name_: Text to set the name for the ProgramType and to be incorporated
            into a unique ProgramType identifier.
        base_program_: An optional ProgramType object that will be used as the
            starting point for the new ProgramType output from this component.
            This can also be text for the name of a ProgramType within the library
            such as that output from the "HB Search Program Types" component.
            If None, a Plenum program type will be used as the base with no loads,
            setpoints, or ventilation requirements assigned.
        _people_: A People object to describe the occupancy of the program. If None,
            no occupancy will be assumed for the program. (Default: None).
        _lighting_: A Lighting object to describe the lighting usage of the program.
            If None, no lighting will be assumed for the program. (Default: None).
        _electric_equip_: An ElectricEquipment object to describe the usage
            of electric equipment within the program. If None, no electric equipment
            will be assumed for the program. (Default: None).
        _gas_equip_: A GasEquipment object to describe the usage of gas equipment
            within the program. If None, no gas equipment will be assumed for
            the program. (Default: None).
        _hot_water_: A ServiceHotWater object to describe the usage of hot water
            within the program. If None, no hot water will be assumed for
            the program. (Default: None).
        _infiltration_: An Infiltration object to describe the outdoor air leakage of
            the program. If None, no infiltration will be assumed for the
            program. (Default: None).
        _ventilation_: A Ventilation object to describe the minimum outdoor air
            requirement of the program. If None, no ventilation requirement will
            be assumed for the program. (Default: None).
        _setpoint_: A Setpoint object to describe the temperature and humidity
            setpoints of the program.  If None, the ProgramType cannot be assigned
            to a Room that is conditioned. (Default: None).

    Returns:
        program: A ProgramType object that can be assigned to Honeybee Rooms in
            order to specify all default schedules and loads on the Room.
"""

ghenv.Component.Name = 'HB ProgramType'
ghenv.Component.NickName = 'ProgramType'
ghenv.Component.Message = '1.2.1'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '0 :: Basic Properties'
ghenv.Component.AdditionalHelpFromDocStrings = '2'

try:  # import the core honeybee dependencies
    from honeybee.typing import clean_and_id_ep_string, clean_ep_string
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:
    from honeybee_energy.programtype import ProgramType
    from honeybee_energy.lib.programtypes import program_type_by_identifier, \
        building_program_type_by_identifier
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # get the base program type
    name = clean_and_id_ep_string('ProgramType') if _name_ is None else \
        clean_ep_string(_name_)
    if base_program_ is None:
        program = ProgramType(name)
    else:
        if isinstance(base_program_, str):
            try:
                base_program_ = building_program_type_by_identifier(base_program_)
            except ValueError:
                base_program_ = program_type_by_identifier(base_program_)
        program = base_program_.duplicate()
        program.identifier = name
    if _name_ is not None:
        program.display_name = _name_

    # go through each input load and assign it to the set
    if _people_ is not None:
        program.people = _people_
    if _lighting_ is not None:
        program.lighting = _lighting_
    if _electric_equip_ is not None:
        program.electric_equipment = _electric_equip_
    if _gas_equip_ is not None:
        program.gas_equipment = _gas_equip_
    if _hot_water_ is not None:
        program.service_hot_water = _hot_water_
    if _infiltration_ is not None:
        program.infiltration = _infiltration_
    if _ventilation_ is not None:
        program.ventilation = _ventilation_
    if _setpoint_ is not None:
        program.setpoint = _setpoint_



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>


"""
Make boundary conditions of Rooms or Faces adiabatic by face type.
-

    Args:
        _hb_objs: Honeybee Faces or Rooms to which adiabatic boundary conditions
            will be assigned.
        exterior_walls_: If True, all exterior walls of the input Rooms or Faces
            will be set to adiabatic. This can also be a list of boolean values
            and different adiabatic values will be assigned based on the cardinal 
            direction, starting with north and moving clockwise.
        roofs_: If True, all exterior roofs of the input Rooms or Faces will be
            set to adiabatic.
        exposed_floors_: If True, all exposed floors of the input Rooms or Faces
            will be set to adiabatic.
        interior_walls_: If True, all interior walls of the input Rooms or Faces
            will be set to adiabatic.
        interior_floors_: If True, all interior floors and ceilings of the input
            Rooms or Faces will be set to adiabatic.
    
    Returns:
        hb_objs: The input honeybee objects with their boundary conditions edited.
"""

ghenv.Component.Name = "HB Adiabatic by Type"
ghenv.Component.NickName = 'AdiabaticByType'
ghenv.Component.Message = '1.2.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '0 :: Basic Properties'
ghenv.Component.AdditionalHelpFromDocStrings = "5"


try:  # import the core honeybee dependencies
    from honeybee.boundarycondition import Outdoors, Ground, Surface, boundary_conditions
    from honeybee.facetype import Wall, RoofCeiling, Floor
    from honeybee.room import Room
    from honeybee.face import Face
    from honeybee.orientation import angles_from_num_orient, face_orient_index
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


def check_type(face, boundary_c, face_type):
    """Check whether a given Face is of a certain type."""
    return isinstance(face.boundary_condition, boundary_c) and \
        isinstance(face.type, face_type)


def check_and_assign_adiabatic_to_face(face):
    """Check if a face if of a relevant type and assign Adiabatic if so."""
    
    # assign the adiabatic property to roofs
    if roofs_ and check_type(face, (Outdoors, Ground), RoofCeiling):
        face.boundary_condition = boundary_conditions.adiabatic
    
    # assign the adiabatic property to exposed floors
    if exposed_floors_ and check_type(face, (Outdoors, Ground), Floor):
        face.boundary_condition = boundary_conditions.adiabatic
    
    # assign the adiabatic property to exposed floors
    if interior_walls_ and check_type(face, Surface, Wall):
        face.boundary_condition = boundary_conditions.adiabatic
    
    # assign the adiabatic property to exposed floors
    if interior_floors_ and check_type(face, Surface, (RoofCeiling, Floor)):
        face.boundary_condition = boundary_conditions.adiabatic


if all_required_inputs(ghenv.Component):
    # duplicate the initial objects
    hb_objs = [obj.duplicate() for obj in _hb_objs]
    
    # assign the adiabatic property to the walls
    if len(exterior_walls_) > 0:
        angles = angles_from_num_orient(len(exterior_walls_))
        for obj in hb_objs:
            if isinstance(obj, Room):
                for face in obj.faces:
                    if check_type(face, (Outdoors, Ground), Wall):
                        orient_i = face_orient_index(face, angles)
                        if orient_i is not None and exterior_walls_[orient_i]:
                            face.boundary_condition = boundary_conditions.adiabatic
            else:  # assume it is a Face
                if check_type(obj, (Outdoors, Ground), Wall):
                    orient_i = face_orient_index(obj, angles)
                    if orient_i is not None and exterior_walls_[orient_i]:
                        obj.boundary_condition = boundary_conditions.adiabatic
    
    # assign the adiabatic property to all of the other face types
    for obj in hb_objs:
        if isinstance(obj, Room):
            for face in obj.faces:
                check_and_assign_adiabatic_to_face(face)
        else:  # assume it is a Face
            check_and_assign_adiabatic_to_face(obj)



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Deconstruct a People object into its constituient properties.
-

    Args:
        _people: A People object to deconstruct.
    
    Returns:
        name: Text string for the people object display name.
         ppl_per_area: A numerical value for the number of people per square
            meter of floor area.
        occupancy_sch: A fractional schedule for the occupancy over the course
            of the year. The fractional values in this schedule get multiplied
            by the _people_per_area to yield a complete occupancy profile.
        activity_sch: A schedule for the activity of the occupants over the
            course of the year. The type limt of this schedule are "Power"
            and the values of the schedule equal to the number of Watts given off
            by an individual person in the room.
"""

ghenv.Component.Name = "HB Deconstruct People"
ghenv.Component.NickName = 'DecnstrPeople'
ghenv.Component.Message = '1.2.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '3 :: Loads'
ghenv.Component.AdditionalHelpFromDocStrings = "0"

try:
    from honeybee_energy.load.people import People
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component) and _people is not None:
    # check the input
    assert isinstance(_people, People), \
        'Expected People object. Got {}.'.format(type(_people))

    # get the properties of the object
    name = _people.display_name
    ppl_per_area = _people.people_per_area
    occupancy_sch = _people.occupancy_schedule
    activity_sch = _people.activity_schedule



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Deconstruct a ProgramType object into its component load objects.
-

    Args:
        _program: A ProgramType object or text for the identifier of a ProgramType
            to be looked up in the program type library.
    
    Returns:
        people: A People object that describes the occupancy of the program. If None,
            no people are assumed to occupy the program.
        lighting: A Lighting object that describes the lighting usage of the program.
            If None, no lights are assumed to be installed.
        electric_equip: An ElectricEquipment object to describe the usage
            of electric equipment within the program. If None, no electric equipment
            is assumed to be installed.
        gas_equip: A GasEquipment object to describe the usage of gas equipment
            within the program. If None, no gas equipment is assumed to be installed.
        hot_water: A ServiceHotWater object to describe the usage of hot water within
            the program. If None, no hot water is be assumed for the program.
        infiltration: An Infiltration object to describe the outdoor air leakage of
            the program. If None, no infiltration is be assumed for the program.
        ventilation: A Ventilation object to describe the minimum outdoor air
            requirement of the program. If None, no ventilation requirement is
            be assumed for the program.
        setpoint: A Setpoint object to describe the temperature and humidity
            setpoints of the program.  If None, the ProgramType cannot be assigned
            to a Room that is conditioned.
"""

ghenv.Component.Name = "HB Deconstruct ProgramType"
ghenv.Component.NickName = 'DeconstrProgram'
ghenv.Component.Message = '1.2.1'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '3 :: Loads'
ghenv.Component.AdditionalHelpFromDocStrings = "1"

try:
    from honeybee_energy.lib.programtypes import program_type_by_identifier, \
        building_program_type_by_identifier
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # get the program from the library if it is a identifier
    if isinstance(_program, str):
        try:
            _program = building_program_type_by_identifier(_program)
        except ValueError:
            _program = program_type_by_identifier(_program)

    # get the components of the program
    people = _program.people
    lighting = _program.lighting
    electric_equip = _program.electric_equipment
    gas_equip = _program.gas_equipment
    hot_water = _program.service_hot_water
    infiltration = _program.infiltration
    ventilation = _program.ventilation
    setpoint = _program.setpoint



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Create a list of interior constructions that can be used to edit or create a
ConstructionSet object.
-

    Args:
        _interior_wall_: A construction object for interior walls (or text for
            the identifier of the construction within the library).
        _ceiling_: A construction object for ceilings (or text for the identifier of
            the construction within the library).
        _interior_floor_: A construction object for interior floors (or text for
            the identifier of the construction within the library).
        _interior_window_: A construction object for all apertures with a Surface
            boundary condition. This can also be text for the identifier of the
            construction within the library.
        _interior_door_: A construction object for all opaque doors with a Surface
            boundary condition. This can also be text for the identifier of the
            construction within the library.
        _int_glass_door_: A construction object for all glass doors with a Surface
            boundary condition. This can also be text for the identifier of the
            construction within the library.
    
    Returns:
        interior_set: A list of interior constructions that can be used to edit
            or create a ConstructionSet object.
"""

ghenv.Component.Name = 'HB Interior Construction Subset'
ghenv.Component.NickName = 'InteriorSubset'
ghenv.Component.Message = '1.2.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '0 :: Basic Properties'
ghenv.Component.AdditionalHelpFromDocStrings = '4'

try:  # import honeybee_energy dependencies
    from honeybee_energy.construction.opaque import OpaqueConstruction
    from honeybee_energy.construction.window import WindowConstruction
    from honeybee_energy.lib.constructions import opaque_construction_by_identifier, \
        window_construction_by_identifier
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))
try:  # import ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import give_warning
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


def opaque_constr(construction, input_name):
    """Get an OpaqueConstrucion from the library if it's a string."""
    if isinstance(construction, str):
        return opaque_construction_by_identifier(construction)
    else:
        assert isinstance(construction, OpaqueConstruction), \
            'Expected OpaqueConstruction for {}. Got {}'.format(
                input_name, type(construction))
    return construction


def window_constr(construction, input_name):
    """Get a WindowConstrucion from the library if it's a string."""
    if isinstance(construction, str):
        return window_construction_by_identifier(construction)
    else:
        assert isinstance(construction, WindowConstruction), \
            'Expected WindowConstruction for {}. Got {}'.format(
                input_name, type(construction))
    return construction


def check_symmetric(constr_type, construction):
    """Give a warning on the component that a construction is asymmetric."""
    if not construction.is_symmetric:   # check whether the construction is symmetric.
        message = 'Input {} is asymmetric (materials in reversed order ' \
            'do not equal those in the current order).\nThis can cause issues if the ' \
            'resulting constr_set is applied across multiple Rooms.\nMaterials: {}'.format(
                constr_type, construction.layers)
        give_warning(ghenv.Component, message)


# go through each input construction
if _interior_wall_ is not None:
    _interior_wall_ = opaque_constr(_interior_wall_, '_interior_wall_')
    check_symmetric('_interior_wall_', _interior_wall_)
if _ceiling_ is not None:
    _ceiling_ = opaque_constr(_ceiling_, '_ceiling_')
if _interior_floor_ is not None:
    _interior_floor_ = opaque_constr(_interior_floor_, '_interior_floor_')
if _interior_window_ is not None:
    _interior_window_ = window_constr(_interior_window_, '_interior_window_')
    check_symmetric('_interior_window_', _interior_window_)
if _interior_door_ is not None:
    _interior_door_ = opaque_constr(_interior_door_, '_interior_door_')
    check_symmetric('_interior_door_', _interior_door_)
if _int_glass_door_ is not None:
    _int_glass_door_ = window_constr(_int_glass_door_, '_int_glass_door_')
    check_symmetric('_int_glass_door_', _int_glass_door_)

# check whether the ceiling has the revered materials of the floor
if _ceiling_ is not None or _interior_floor_ is not None:
    if _ceiling_ is None or _interior_floor_ is  None or \
            _ceiling_.layers != list(reversed(_interior_floor_.layers)):
        warn = '_ceiling_ does not have materials in reversed ' \
            ' order of the _interior_floor_.\nThis can cause issues if the ' \
            'resulting constr_set is applied across multiple Rooms.'
        give_warning(ghenv.Component, warn)
        print(warn)

# return the final list from the component
interior_set = [_interior_wall_, _ceiling_, _interior_floor_, _interior_window_,
                _interior_door_, _int_glass_door_]



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Create a Ventilation Control object to dictate the temperature setpoints and
schedule for ventilative cooling (eg. opening windows).
_
Note the all of the default setpoints of this object are set to always perform
ventilative cooling such that one can individually decide which setpoints
are relevant to a given ventilation strategy.
-

    Args:
        min_in_temp_: A number between -100 and 100 for the minimum indoor
            temperature at which to ventilate in Celsius. Typically,
            this variable is used to initiate ventilation with values
            around room temperature above which the windows will open
            (eg. 22 C). (Default: -100 C).
        max_in_temp_: A number between -100 and 100 for the maximum indoor
            temperature at which to ventilate in Celsius. This can be
            used to set a maximum temperature at which point ventilation is
            stopped and a cooling system is turned on. (Default: 100 C).
        min_out_temp_: A number between -100 and 100 for the minimum outdoor
            temperature at which to ventilate in Celsius. This can be used
            to ensure ventilative cooling doesn't happen during the winter even
            if the Room is above the min_in_temp. (Default: -100 C).
        max_out_temp_: A number between -100 and 100 for the maximum outdoor
            temperature at which to ventilate in Celsius. This can be used
            to set a limit for when it is considered too hot outside for
            ventilative cooling. (Default: 100).
        delta_temp_: A number between -100 and 100 for the temperature differential
            in Celsius between indoor and outdoor below which ventilation
            is shut off.  This should usually be a positive number so that
            ventilation only occurs when the outdoors is cooler than the indoors.
            Negative numbers indicate how much hotter the outdoors can be than
            the indoors before ventilation is stopped. (Default: -100).
        _schedule_: An optional schedule for the ventilation over the course of
            the year. This can also be the name of a schedule to be looked up
            in the standards library. Note that this is applied on top of any
            setpoints. The type of this schedule should be On/Off and values
            should be either 0 (no possibility of ventilation) or 1 (ventilation
            possible). (Default: "Always On")

    Returns:
        vent_cntrl: HBZones with their airflow modified.
"""

ghenv.Component.Name = 'HB Ventilation Control'
ghenv.Component.NickName = 'VentControl'
ghenv.Component.Message = '1.2.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '3 :: Loads'
ghenv.Component.AdditionalHelpFromDocStrings = '4'

try:
    from honeybee_energy.ventcool.control import VentilationControl
    from honeybee_energy.lib.schedules import schedule_by_identifier
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))


# set default values
min_in_temp_ = -100 if min_in_temp_ is None else min_in_temp_
max_in_temp_ = 100 if max_in_temp_ is None else max_in_temp_
min_out_temp_ = -100 if min_out_temp_ is None else min_out_temp_
max_out_temp_ = 100 if max_out_temp_ is None else max_out_temp_
delta_temp_ = -100 if delta_temp_ is None else delta_temp_

# get the schedule if it's just an identifier
if isinstance(_schedule_, str):
    _schedule_ = schedule_by_identifier(_schedule_)

# create the VentilationControl object
vent_cntrl = VentilationControl(
    min_in_temp_, max_in_temp_, min_out_temp_, max_out_temp_, delta_temp_, _schedule_)



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Create an Infiltration object that can be used to create a ProgramType or be
assigned directly to a Room.
-

    Args:
        _name_: Text to set the name for the Infiltration and to be incorporated
            into a unique Infiltration identifier. If None, a unique name will
            be generated.
        _flow_per_ext_area: A numerical value for the intensity of infiltration
            in m3/s per square meter of exterior surface area. Typical values for
            this property are as follows (note all values are at typical building
            pressures of ~4 Pa):
                * 0.0001 (m3/s per m2 facade) - Tight building
                * 0.0003 (m3/s per m2 facade) - Average building
                * 0.0006 (m3/s per m2 facade) - Leaky building
        _schedule_: A fractional schedule for the infiltration over the course
            of the year. The fractional values will get multiplied by the
            flow_per_exterior_area to yield a complete infiltration profile.
            Default: 'Always On'

    Returns:
        infil: An Infiltration object that can be used to create a ProgramType or
            be assigned directly to a Room.
"""

ghenv.Component.Name = 'HB Infiltration'
ghenv.Component.NickName = 'Infiltration'
ghenv.Component.Message = '1.2.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '3 :: Loads'
ghenv.Component.AdditionalHelpFromDocStrings = '3'

try:  # import the core honeybee dependencies
    from honeybee.typing import clean_and_id_ep_string, clean_ep_string
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:
    from honeybee_energy.load.infiltration import Infiltration
    from honeybee_energy.lib.schedules import schedule_by_identifier
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # make a default Infiltration name if none is provided
    name = clean_and_id_ep_string('Infiltration') if _name_ is None else \
        clean_ep_string(_name_)

    # get the schedule
    _schedule_ = _schedule_ if _schedule_ is not None else 'Always On'
    if isinstance(_schedule_, str):
        _schedule_ = schedule_by_identifier(_schedule_)

    # create the Infiltration object
    infil = Infiltration(name, _flow_per_ext_area, _schedule_)
    if _name_ is not None:
        infil.display_name = _name_



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Deconstruct a Setpoint object into its constituient properties.
-

    Args:
        _setpoint: A Setpoint object to be deconstructed.
    
    Returns:
        name: Text string for the setpoint display name.
         heating_sch: A temperature schedule for the heating setpoint.
            The type limit of this schedule should be temperature and the
            values should be the temperature setpoint in degrees Celcius.
        cooling_sch: A temperature schedule for the cooling setpoint.
            The type limit of this schedule should be temperature and the
            values should be the temperature setpoint in degrees Celcius.
        humid_setpt: A numerical value between 0 and 100 for the relative humidity
            humidifying setpoint [%]. This value will be constant throughout the
            year. If None, no humidification will occur.
        dehumid_setpt: A numerical value between 0 and 100 for the relative humidity
            dehumidifying setpoint [%]. This value will be constant throughout the
            year. If None, no dehumidification will occur beyond that which is needed
            to create air at the cooling supply temperature.
"""

ghenv.Component.Name = "HB Deconstruct Setpoint"
ghenv.Component.NickName = 'DecnstrSetpoint'
ghenv.Component.Message = '1.2.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '3 :: Loads'
ghenv.Component.AdditionalHelpFromDocStrings = "0"

try:
    from honeybee_energy.load.setpoint import Setpoint
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component) and _setpoint is not None:
    # check the input
    assert isinstance(_setpoint, Setpoint), \
        'Expected Setpoint object. Got {}.'.format(type(_setpoint))

    # get the properties of the object
    name = _setpoint.display_name
    heating_sch = _setpoint.heating_schedule
    cooling_sch = _setpoint.cooling_schedule
    humid_setpt = _setpoint.humidifying_setpoint
    dehumid_setpt = _setpoint.dehumidifying_setpoint



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Normalize Zone-level data collections from an energy simulation by the by the
floor area of the corresponding honeybee Rooms.
-

    Args:
        _data: A list of HourlyContinuousCollections of the same data type, which
            will be normalized by room floor area. Data collections can be of any class
            (eg. MonthlyCollection, DailyCollection) but they should originate
            from an energy simulation sql (with header metadata that has 'Zone' or
            'System' keys). These keys will be used to match the data in the collections
            to the input rooms.
        _model: An array of honeybee Rooms or a honeybee Model, which will
            be matched to the data collections. The length of these Rooms does
            not have to match the data collections and this object will only
            output collections for rooms that are found to be matching.

    Returns:
        total_data: The total results normalized by the floor area of all connected
            rooms. This accounts for the fact that some rooms have more floor
            area (or have a multiplier) and therefore get a greater weighting.
        room_data: The results normalized by the floor area of each individual room.
"""

ghenv.Component.Name = "HB Normalize by Floor Area"
ghenv.Component.NickName = 'NormByFlr'
ghenv.Component.Message = '1.2.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '6 :: Result'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:
    from ladybug.header import Header
    from ladybug.datacollection import HourlyContinuousCollection
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from honeybee.model import Model
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:
    from honeybee_energy.result.match import match_rooms_to_data
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs, give_warning
    from ladybug_rhino.config import conversion_to_meters
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # extract any rooms from input Models and convert geo to meters if necessary
    rooms = []
    for hb_obj in _model:
        if isinstance(hb_obj, Model):
            rooms.extend(hb_obj.rooms)
        else:
            rooms.append(hb_obj)
    m_convert = conversion_to_meters()
    if m_convert != 1:  # duplicate and scale all objects to meters
        rooms = [room.duplicate() for room in rooms]
        [room.scale(m_convert) for room in rooms]

    # match the data with the rooms
    match_tups = match_rooms_to_data(_data, rooms)

    # divide the individual data collections by floor area
    total_area = 0
    room_data = []
    for tup in match_tups:
        total_flr_area = tup[0].floor_area * tup[2]  # includes effect of multiplier
        total_area += total_flr_area
        try:
            room_data.append(tup[1].normalize_by_area(total_flr_area, 'm2'))
        except ZeroDivisionError:  # no floor area; not normalizable
            pass

    # sum all collections together and normalize them by the total
    if len(match_tups) != 0:
        summed_vals = [val for val in match_tups[0][1]]
        for data_i in match_tups[1:]:
            for i, val in enumerate(data_i[1]):
                summed_vals[i] += val
    else:  # just assume all of the data corresponds with all input rooms
        summed_vals = [0 for val in _data[0]]
        total_area = sum(room.floor_area * room.multiplier for room in rooms)
        for d in _data:
            for i, val in enumerate(d):
                summed_vals[i] += val
    try:
        total_data = _data[0].duplicate()
        total_data.values = summed_vals
        total_data = total_data.normalize_by_area(total_area, 'm2')
        total_data.header.metadata = {'type': _data[0].header.metadata['type']}
    except ZeroDivisionError:  # no floors in the model
        give_warning(ghenv.Component, 'No floors were found in the input _model.')



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Compute spatially-resolved Universal Thermal Climate Index (UTCI) and heat/cold
stress conditions an EPW and Honeybee model.
_
This recipe uses EnergyPlus to obtain longwave radiant temperatures and indoor air
temperatures. The outdoor air temperature and air speed are taken directly from
the EPW. A Radiance-based, enhanced 2-phase method is used for shortwave MRT
calculations, which uses an accurate direct sun calculation with precise solar
positions.
_
The energy properties of the model geometry are what determine the outcome of the
simulation, though the model's Radiance sensor grids are what determine where
the comfort mapping occurs.
-

    Args:
        _model: A Honeybee Model for which UTCI comfort will be mapped. Note that
            this model should have radiance grids assigned to it in order
            to produce meaningful results.
        _epw: Path to an EPW weather file to be used for the comfort map simulation.
        _ddy: Path to a DDY file with design days to be used for the initial sizing
            calculation of the energy simulation.
        north_: A number between -360 and 360 for the counterclockwise difference
            between the North and the positive Y-axis in degrees. This can
            also be Vector for the direction to North. (Default: 0).
        run_period_: An AnalysisPeriod to set the start and end dates of the simulation.
            If None, the simulation will be annual.
        sensor_count_: Integer for the maximum number of sensor grid points per
            parallel execution. (Default: 200).
        _wind_speed_: A single number for meteorological wind speed in m/s or an hourly
            data collection of wind speeds that align with the input run_period_.
            This will be used for all indoor comfort evaluation. Note that the
            EPW wind speed will be used for any outdoor sensors. (Default: 0.5).
        solar_body_par_: Optional solar body parameters from the "LB Solar Body Parameters"
            object to specify the properties of the human geometry assumed in the
            shortwave MRT calculation. The default assumes average skin/clothing
            absorptivity and a human subject always has their back to the sun
            at a 45-degree angle (SHARP = 135).
        radiance_par_: Text for the radiance parameters to be used for ray
            tracing. (Default: -ab 2 -ad 5000 -lw 2e-05).
        run_settings_: Settings from the "HB Recipe Settings" component that specify
            how the recipe should be run. This can also be a text string of
            recipe settings.
        _run: Set to True to run the recipe and get results. This input can also be
            the integer "2" to run the recipe silently.

    Returns:
        report: Reports, errors, warnings, etc.
        utci: A folder containing CSV maps of Universal Thermal Climate Index (UTCI)
            temperatures for each sensor grid at each time step of the analysis.
            This can be connected to the "HB Read Thermal Matrix" component to
            parse detailed results into Grasshopper. Values are in Celsius.
        condition: A folder containing CSV maps of comfort conditions for each sensor
            grid at each time step of the analysis. This can be connected to the
            "HB Read Thermal Matrix" component to parse detailed results into
            Grasshopper. Values are as follows.
                * -1 = unacceptably cold conditions
                *  0 = neutral (comfortable) conditions
                * +1 = unacceptably hot conditions
        category: A folder containing CSV maps of the heat/cold stress categories for
            each sensor grid at each time step of the analysis. This can be connected
            to the "HB Read Thermal Matrix" component to parse detailed results
            into Grasshopper. This can be used to understand not just whether
            conditions are acceptable but how uncomfortably hot or cold they
            are. Values indicate the following.
                * -5 = extreme cold stress
                * -4 = very strong cold stress
                * -3 = strong cold stress
                * -2 = moderate cold stress
                * -1 = slight cold stress
                *  0 = no thermal stress
                * +1 = slight heat stress
                * +2 = moderate heat stress
                * +3 = strong heat stress
                * +4 = very strong heat stress
                * +5 = extreme heat stress
        TCP: Lists of values between 0 and 100 for the Thermal Comfort Percent (TCP).
            These can be plugged into the "LB Spatial Heatmap" component along
            with meshes of the sensor grids to visualize spatial thermal comfort.
            TCP is the percentage of occupied time where thermal conditions are
            acceptable/comfortable. Occupied hours are determined from the
            occuppancy schedules of each room (any time where the occupancy
            schedule is >= 0.1 will be considered occupied). Outdoor sensors
            are considered occupied at all times. More custom TCP studies can
            be done by post-processing the condition results.
        HSP: Lists of values between 0 and 100 for the Heat Sensation Percent (HSP).
            These can be plugged into the "LB Spatial Heatmap" component along with
            meshes of the sensor grids to visualize uncomfortably hot locations.
            HSP is the percentage of occupied time where thermal conditions are
            hotter than what is considered acceptable/comfortable. Occupied hours
            are determined from the occuppancy schedules of each room (any time
            where the occupancy schedule is >= 0.1 will be considered occupied).
            Outdoor sensors are considered occupied at all times. More custom HSP
            studies can be done by post-processing the condition results.
        CSP: Lists of values between 0 and 100 for the Cold Sensation Percent (CSP).
            These can be plugged into the "LB Spatial Heatmap" component along with
            meshes of the sensor grids to visualize uncomfortably cold locations.
            CSP is the percentage of occupied time where thermal conditions are
            colder than what is considered acceptable/comfortable. Occupied hours
            are determined from the occuppancy schedules of each room (any time
            where the occupancy schedule is >= 0.1 will be considered occupied).
            Outdoor sensors are considered occupied at all times. More custom CSP
            studies can be done by post-processing the condition results.
"""

ghenv.Component.Name = 'HB UTCI Comfort Map'
ghenv.Component.NickName = 'UTCIMap'
ghenv.Component.Message = '1.2.3'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '7 :: Thermal Map'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

try:
    from lbt_recipes.recipe import Recipe
except ImportError as e:
    raise ImportError('\nFailed to import lbt_recipes:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs, recipe_result
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component) and _run:
    # create the recipe and set the input arguments
    recipe = Recipe('utci_comfort_map')
    recipe.input_value_by_name('model', _model)
    recipe.input_value_by_name('epw', _epw)
    recipe.input_value_by_name('ddy', _ddy)
    recipe.input_value_by_name('north', north_)
    recipe.input_value_by_name('run-period', run_period_)
    recipe.input_value_by_name('sensor-count', sensor_count_)
    recipe.input_value_by_name('wind-speed', _wind_speed_)
    recipe.input_value_by_name('solarcal-parameters', solar_body_par_)
    recipe.input_value_by_name('radiance-parameters', radiance_par_)

    # run the recipe
    silent = True if _run > 1 else False
    project_folder = recipe.run(
        run_settings_, radiance_check=True, openstudio_check=True, silent=silent)

    # load the results
    try:
        utci = recipe_result(recipe.output_value_by_name('utci', project_folder))
        condition = recipe_result(recipe.output_value_by_name('condition', project_folder))
        category = recipe_result(recipe.output_value_by_name('category', project_folder))
        TCP = recipe_result(recipe.output_value_by_name('tcp', project_folder))
        HSP = recipe_result(recipe.output_value_by_name('hsp', project_folder))
        CSP = recipe_result(recipe.output_value_by_name('csp', project_folder))
    except Exception:
        raise Exception(recipe.failure_message(project_folder))



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Write a honeybee Model to an OSM file (OpenStudio Model), which can then be translated
to an IDF file and then run through EnergyPlus.

-
    Args:
        _model: A honeybee model object possessing all geometry and corresponding
            energy simulation properties.
        _epw_file: Path to an .epw file on this computer as a text string.
        _sim_par_: A honeybee Energy SimulationParameter object that describes all
            of the setting for the simulation. If None, some default simulation
            parameters will automatically be used.
        measures_: An optional list of measures to apply to the OpenStudio model
            upon export. Use the "HB Load Measure" component to load a measure
            into Grasshopper and assign input arguments. Measures can be
            downloaded from the NREL Building Components Library (BCL) at
            (https://bcl.nrel.gov/).
        add_str_: THIS OPTION IS JUST FOR ADVANCED USERS OF ENERGYPLUS.
            You can input additional text strings here that you would like
            written into the IDF.  The input here should be complete EnergyPlus
            objects as a single string following the IDF format. This input ca
            be used to write objects into the IDF that are not currently supported
            by Honeybee.
        _folder_: An optional folder on this computer, into which the IDF and result
            files will be written.  NOTE THAT DIRECTORIES INPUT HERE SHOULD NOT HAVE
            ANY SPACES OR UNDERSCORES IN THE FILE PATH.
        _write: Set to "True" to write out the honeybee jsons (containing the Honeybee
            Model and Simulation Parameters) and write the OpenStudio Workflow
            (.osw) file with instructions for executing the simulation.
        run_: Set to "True" to translate the Honeybee jsons to an OpenStudio Model
            (.osm) and EnergyPlus Input Data File (.idf) and then simulate the
            .idf in EnergyPlus. This will ensure that all result files appear
            in their respective outputs from this component.
            _
            This input can also be the integer "2", which will only translate the
            honeybee jsons to an osm and idf format without running the model
            through EnergyPlus.
            _
            It can also be the integer "3", which will run the whole translation
            and simulation silently (without any batch windows).

    Returns:
        report: Check here to see a report of the EnergyPlus run.
        jsons: The file paths to the honeybee JSON files that describe the Model and
            SimulationParameter. These will be translated to an OpenStudio model.
        osw: File path to the OpenStudio Workflow JSON on this machine. This workflow
            is executed using the OpenStudio command line interface (CLI) and
            it includes measures to translate the Honeybee model JSON as well
            as any other connected measures_.
        osm: The file path to the OpenStudio Model (OSM) that has been generated
            on this computer.
        idf: The file path of the EnergyPlus Input Data File (IDF) that has been
            generated on this computer.
        sql: The file path of the SQL result file that has been generated on this
            computer. This will be None unless run_ is set to True.
        zsz: Path to a .csv file containing detailed zone load information recorded
            over the course of the design days. This will be None unless run_ is
            set to True.
        rdd: The file path of the Result Data Dictionary (.rdd) file that is
            generated after running the file through EnergyPlus.  This file
            contains all possible outputs that can be requested from the EnergyPlus
            model. Use the "HB Read Result Dictionary" component to see what outputs
            can be requested.
        html: The HTML file path containing all requested Summary Reports.
"""

ghenv.Component.Name = 'HB Model to OSM'
ghenv.Component.NickName = 'ModelToOSM'
ghenv.Component.Message = '1.2.3'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '5 :: Simulate'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

import os
import json

try:
    from ladybug.futil import preparedir, nukedir
    from ladybug.epw import EPW
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    import honeybee.config as hb_config
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:
    from honeybee_energy.simulation.parameter import SimulationParameter
    from honeybee_energy.run import to_openstudio_osw, run_osw, run_idf, \
        output_energyplus_files
    from honeybee_energy.result.err import Err
    from honeybee_energy.result.osw import OSW
    from honeybee_energy.config import folders as energy_folders
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from lbt_recipes.version import check_openstudio_version
except ImportError as e:
    raise ImportError('\nFailed to import lbt_recipes:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs, give_warning
    from ladybug_rhino.config import units_system
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


def orphaned_warning(object_type):
    """Generate an error message for orphaned Faces, Apertures, or Doors."""
    return 'Input _model contains orphaned {}s. These are not permitted in ' \
        'Models for energy simulation.\nIf you have geometry that is not a ' \
        'part of a Room boundary that you want included in the energy simulation, ' \
        'it should be added as shades.'.format(object_type)


if all_required_inputs(ghenv.Component) and _write:
    # check the presence of openstudio and check that the version is compatible
    check_openstudio_version()

    # process the simulation parameters
    if _sim_par_ is None:
        _sim_par_ = SimulationParameter()
        _sim_par_.output.add_zone_energy_use()
        _sim_par_.output.add_hvac_energy_use()
    else:
        _sim_par_ = _sim_par_.duplicate()  # ensure input is not edited

    # assign design days from the DDY next to the EPW if there are None
    if len(_sim_par_.sizing_parameter.design_days) == 0:
        msg = None
        folder, epw_file_name = os.path.split(_epw_file)
        ddy_file = os.path.join(folder, epw_file_name.replace('.epw', '.ddy'))
        if os.path.isfile(ddy_file):
            try:
                _sim_par_.sizing_parameter.add_from_ddy_996_004(ddy_file)
            except AssertionError:
                msg = 'No ddy_file_ was input into the _sim_par_ sizing ' \
                    'parameters\n and no design days were found in the .ddy file '\
                    'next to the _epw_file.'
        else:
             msg = 'No ddy_file_ was input into the _sim_par_ sizing parameters\n' \
                'and no .ddy file was found next to the _epw_file.'
        if msg is not None:
            epw_obj = EPW(_epw_file)
            des_days = [epw_obj.approximate_design_day('WinterDesignDay'),
                        epw_obj.approximate_design_day('SummerDesignDay')]
            _sim_par_.sizing_parameter.design_days = des_days
            msg = msg + '\nDesign days were generated from the input _epw_file but this ' \
                '\nis not as accurate as design days from DDYs distributed with the EPW.'
            give_warning(ghenv.Component, msg)
            print msg

    # process the simulation folder name and the directory
    _folder_ = hb_config.folders.default_simulation_folder if _folder_ is None else _folder_
    directory = os.path.join(_folder_, _model.identifier, 'OpenStudio')

    # check the model to be sure there are no orphaned faces, apertures, or doors
    assert len(_model.orphaned_faces) == 0, orphaned_warning('Face')
    assert len(_model.orphaned_apertures) == 0, orphaned_warning('Aperture')
    assert len(_model.orphaned_doors) == 0, orphaned_warning('Door')

    # duplicate model to avoid mutating it as we edit it for energy simulation
    _model = _model.duplicate()
    # remove colinear vertices using the Model tolerance to avoid E+ tolerance issues
    for room in _model.rooms:
        room.remove_colinear_vertices_envelope(_model.tolerance)
    # auto-assign stories if there are none since most OpenStudio measures need these
    if len(_model.stories) == 0:
        _model.assign_stories_by_floor_height()
    # scale the model if the units are not meters
    if _model.units != 'Meters':
        _model.convert_to_units('Meters')

    # delete any existing files in the directory and prepare it for simulation
    nukedir(directory, True)
    preparedir(directory)
    sch_directory = os.path.join(directory, 'schedules')
    preparedir(sch_directory)

    # write the model parameter JSONs
    model_dict = _model.to_dict(triangulate_sub_faces=True)
    model_json = os.path.join(directory, '{}.hbjson'.format(_model.identifier))
    with open(model_json, 'w') as fp:
        json.dump(model_dict, fp)

    # write the simulation parameter JSONs
    sim_par_dict = _sim_par_.to_dict()
    sim_par_json = os.path.join(directory, 'simulation_parameter.json')
    with open(sim_par_json, 'w') as fp:
        json.dump(sim_par_dict, fp)

    # process any measures input to the component
    measures = None if len(measures_) == 0 or measures_[0] is None else measures_
    no_report_meas = True if measures is None else \
        all(meas.type != 'ReportingMeasure' for meas in measures)

    # collect the two jsons for output and write out the osw file
    jsons = [model_json, sim_par_json]
    osw = to_openstudio_osw(
        directory, model_json, sim_par_json, additional_measures=measures,
        epw_file=_epw_file, schedule_directory=sch_directory)

    # run the measure to translate the model JSON to an openstudio measure
    silent = True if run_ == 3 else False
    if run_ > 0 and not no_report_meas:  # everything must run with OS CLI
        osm, idf = run_osw(osw, measures_only=False, silent=silent)
        sql, zsz, rdd, html, err = output_energyplus_files(os.path.dirname(idf))
    elif run_ > 0:  # no reporting measure; simulate separately from measure application
        osm, idf = run_osw(osw, silent=silent)
        # process the additional strings
        if add_str_ != [] and add_str_[0] is not None and idf is not None:
            add_str = '\n'.join(add_str_)
            with open(idf, "a") as idf_file:
                idf_file.write(add_str)
        if idf is None:  # measures failed to run correctly; parse out.osw
            log_osw = OSW(os.path.join(directory, 'out.osw'))
            for error, tb in zip(log_osw.errors, log_osw.error_tracebacks):
                if 'Cannot create a surface' in error:
                    error = 'Your Rhino Model units system is: {}\n{}'.format(
                        units_system(), error)
                print(tb)
                raise Exception(error)
        if run_ in (1, 3):  # run the resulting idf throught EnergyPlus
            sql, zsz, rdd, html, err = run_idf(idf, _epw_file, silent=silent)

    # parse the error log and report any warnings
    if run_ in (1, 3) and err is not None:
        err_obj = Err(err)
        print(err_obj.file_contents)
        for warn in err_obj.severe_errors:
            give_warning(ghenv.Component, warn)
        for error in err_obj.fatal_errors:
            raise Exception(error)



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Apply simple daylight controls to Rooms.
_
Such simple controls will dim the lights in the energy simulation according to
whether the illuminance at a senor location is at a target illuminance setpoint.
The method used to estimate illuiminance is fairly simple and, for more detailed
control over the parameters used to compute illuminance, HB-Radiance should
be used.
-

    Args:
        _rooms: Honeybee Rooms to which simple daylight controls should be assigned.
        _sensor_points_: A list of point objects that align with the input _rooms and
            assign the position of the daylight sensor within the Room.
            This point should lie within the Room volume and a warning will
            be thrown and no daylight controls assigned for any point that
            lies outside the corresponding room. If unspecified, the
            sensor will be assigned to the center of the room at 0.8 meters
            above the floor. Note that such a center point might lie outside
            rooms that are significantly concave and no daylight controls
            will be assigned to these rooms in this case.
        _ill_setpoint_: A number for the illuminance setpoint in lux beyond which
            electric lights are dimmed if there is sufficient daylight.
            Some common setpoints are listed below. (Default: 300 lux).
            -
            50 lux - Corridors and hallways.
            150 lux - Computer work spaces (screens provide illumination).
            300 lux - Paper work spaces (reading from surfaces that need illumination).
            500 lux - Retail spaces or museums illuminating merchandise/artifacts.
            1000 lux - Operating rooms and workshops where light is needed for safety.

        _control_fract_: A number between 0 and 1 that represents the fraction of
            the Room lights that are dimmed when the illuminance at the sensor
            position is at the specified illuminance. 1 indicates that all lights are
            dim-able while 0 indicates that no lights are dim-able. Deeper rooms
            should have lower control fractions to account for the face that the
            lights in the back of the space do not dim in response to suitable
            daylight at the front of the room. (Default: 1).
        _min_power_in_: A number between 0 and 1 for the the lowest power the lighting
            system can dim down to, expressed as a fraction of maximum
            input power. (Default: 0.3).
        _min_light_out_: A number between 0 and 1 the lowest lighting output the lighting
            system can dim down to, expressed as a fraction of maximum
            light output. (Default: 0.2).
        off_at_min_: Boolean to note whether lights should switch off completely when
            they get to the minimum power input. (Default: False).

    Returns:
        report: Reports, errors, warnings, etc.
        rooms: The input Rooms with simple daylight controls assigned to them.
"""

ghenv.Component.Name = 'HB Apply Daylight Control'
ghenv.Component.NickName = 'DaylightControl'
ghenv.Component.Message = '1.2.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '3 :: Loads'
ghenv.Component.AdditionalHelpFromDocStrings = '2'

try:  # import the honeybee-energy extension
    from honeybee_energy.load.daylight import DaylightingControl
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from ladybug_rhino.togeometry import to_point3d
    from ladybug_rhino.config import conversion_to_meters
    from ladybug_rhino.grasshopper import all_required_inputs, longest_list, \
        give_warning
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # duplicate the initial objects
    rooms = [room.duplicate() for room in _rooms]

    # set default values and perform checks
    dist_from_floor = 0.8 / conversion_to_meters()
    if len(_sensor_points_) != 0:
        assert len(_sensor_points_) == len(_rooms), 'Number of sensor points ({}) ' \
            'must align exactly with the number of rooms ({}).'.format(
                len(_sensor_points_), len(_rooms))
    _ill_setpoint_ = [300] if len(_ill_setpoint_) == 0 else _ill_setpoint_
    _control_fract_ = [1] if len(_control_fract_) == 0 else _control_fract_
    _min_power_in_ = [0.3] if len(_min_power_in_) == 0 else _min_power_in_
    _min_light_out_ = [0.2] if len(_min_light_out_) == 0 else _min_light_out_
    off_at_min_ = [False] if len(off_at_min_) == 0 else off_at_min_

    # loop through the rooms and assign daylight sensors
    unassigned_rooms = []
    if len(_sensor_points_) == 0:
        for i, room in enumerate(rooms):
            dl_control = room.properties.energy.add_daylight_control_to_center(
                dist_from_floor, longest_list(_ill_setpoint_, i),
                longest_list(_control_fract_, i), longest_list(_min_power_in_, i),
                longest_list(_min_light_out_, i), longest_list(off_at_min_, i))
            if dl_control is None:
                unassigned_rooms.append(room.display_name)
    else:
        for i, room in enumerate(rooms):
            sensor_pt = to_point3d(_sensor_points_[i])
            if room.geometry.is_point_inside(sensor_pt):
                dl_control = DaylightingControl(
                    sensor_pt, longest_list(_ill_setpoint_, i),
                    longest_list(_control_fract_, i), longest_list(_min_power_in_, i),
                    longest_list(_min_light_out_, i), longest_list(off_at_min_, i))
                room.properties.energy.daylighting_control = dl_control
            else:
                unassigned_rooms.append(room.display_name)

    # give a warning about any rooms to which a sensor could not be assinged
    for room in unassigned_rooms:
        msg = 'Sensor point for room "{}" does not lie within the room volume.\n' \
            'No daylight sensors have been added to this room.'.format(room)
        print(msg)
        give_warning(ghenv.Component, msg)



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>


"""
Deconstruct a construction set into its constituient interior constructions.
-

    Args:
        _constr_set: A construction set to be deconstructed. This can also be
            text for a construction set to be looked up in the construction
            set library.

    Returns:
        interior_wall: A construction object for the set's interior walls.
        ceiling: A construction object for the set's interior roofs.
        interior_floor: A construction object for the set's interior floors.
        interior_window: A construction object for all apertures with a Surface
            boundary condition.
        interior_door: A construction object for all opaque doors with a Surface
            boundary condition.
        int_glass_door: A construction object for all glass doors with a Surface
            boundary condition.
"""

ghenv.Component.Name = "HB Deconstruct ConstructionSet Interior"
ghenv.Component.NickName = 'DecnstrConstrSetInt'
ghenv.Component.Message = '1.2.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = "1 :: Constructions"
ghenv.Component.AdditionalHelpFromDocStrings = "2"


try:  # import the honeybee-energy dependencies
    from honeybee_energy.constructionset import ConstructionSet
    from honeybee_energy.lib.constructionsets import construction_set_by_identifier
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))
try:  # import ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # check the input
    if isinstance(_constr_set, str):
        _constr_set = construction_set_by_identifier(_constr_set)
    else:
        assert isinstance(_constr_set, ConstructionSet), \
            'Expected ConstructionSet. Got {}.'.format(type(_constr_set))
    
    interior_wall = _constr_set.wall_set.interior_construction
    ceiling = _constr_set.roof_ceiling_set.interior_construction
    interior_floor = _constr_set.floor_set.interior_construction
    interior_window = _constr_set.aperture_set.interior_construction
    interior_door = _constr_set.door_set.interior_construction
    int_glass_door = _constr_set.door_set.interior_glass_construction



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Search for available Materials within the honeybee energy standards library.
-

    Args:
        keywords_: Optional keywords to be used to narrow down the output list of
            materials. If nothing is input here, all available materials
            will be output.
        join_words_: If False or None, this component will automatically split
            any strings of multiple keywords (spearated by spaces) into separate
            keywords for searching. This results in a greater liklihood of
            finding an item in the search but it may not be appropropriate for
            all cases. You may want to set it to True when you are searching for
            a specific phrase that includes spaces. Default: False.
    
    Returns:
        opaque_mats: A list of opaque materials within the honeybee energy
            standards library (filtered by keywords_ if they are input).
        window_mats: A list of window materials within the honeybee energy
            standards library (filtered by keywords_ if they are input).
"""

ghenv.Component.Name = "HB Search Materials"
ghenv.Component.NickName = 'SearchMats'
ghenv.Component.Message = '1.2.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = "1 :: Constructions"
ghenv.Component.AdditionalHelpFromDocStrings = "1"

try:  # import the honeybee-core dependencies
    from honeybee.search import filter_array_by_keywords
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:  # import the honeybee-energy dependencies
    from honeybee_energy.lib.materials import OPAQUE_MATERIALS
    from honeybee_energy.lib.materials import WINDOW_MATERIALS
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))


if len(keywords_) == 0:
    opaque_mats = sorted(OPAQUE_MATERIALS)
    window_mats = sorted(WINDOW_MATERIALS)
else:
    split_words = True if join_words_ is None else not join_words_
    opaque_mats = sorted(filter_array_by_keywords(OPAQUE_MATERIALS, keywords_, split_words))
    window_mats = sorted(filter_array_by_keywords(WINDOW_MATERIALS, keywords_, split_words))
